"""Compute functions for quasisymmetry objectives.

Notes
-----
Some quantities require additional work to compute at the magnetic axis.
A Python lambda function is used to lazily compute the magnetic axis limits
of these quantities. These lambda functions are evaluated only when the
computational grid has a node on the magnetic axis to avoid potentially
expensive computations.
"""

import numpy as np

from desc.backend import jnp, put, sign

from .data_index import register_compute_fun
from .utils import cross, dot


@register_compute_fun(
    name="B_theta_mn",
    label="B_{\\theta, m, n}",
    units="T \\cdot m}",
    units_long="Tesla * meters",
    description="Fourier coefficients for covariant poloidal component of "
    + "magnetic field",
    dim=1,
    params=[],
    transforms={"B": [[0, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=["B_theta"],
)
def _B_theta_mn(params, transforms, profiles, data, **kwargs):
    data["B_theta_mn"] = transforms["B"].fit(data["B_theta"])
    return data


@register_compute_fun(
    name="B_zeta_mn",
    label="B_{\\zeta, m, n}",
    units="T \\cdot m}",
    units_long="Tesla * meters",
    description="Fourier coefficients for covariant toroidal component of "
    + "magnetic field",
    dim=1,
    params=[],
    transforms={"B": [[0, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=["B_zeta"],
)
def _B_zeta_mn(params, transforms, profiles, data, **kwargs):
    data["B_zeta_mn"] = transforms["B"].fit(data["B_zeta"])
    return data


@register_compute_fun(
    name="w_Boozer_mn",
    label="w_{\\mathrm{Boozer},m,n}",
    units="T \\cdot m",
    units_long="Tesla * meters",
    description="RHS of eq 10 in Hirshman 1995 'Transformation from VMEC to "
    + "Boozer Coordinates'",
    dim=1,
    params=[],
    transforms={"w": [[0, 0, 0]], "B": [[0, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=["B_theta_mn", "B_zeta_mn"],
)
def _w_mn(params, transforms, profiles, data, **kwargs):
    w_mn = jnp.zeros((transforms["w"].basis.num_modes,))
    Bm = transforms["B"].basis.modes[:, 1]
    Bn = transforms["B"].basis.modes[:, 2]
    wm = transforms["w"].basis.modes[:, 1]
    wn = transforms["w"].basis.modes[:, 2]
    NFP = transforms["w"].basis.NFP
    # indices of matching modes in w and B bases
    # need to use np instead of jnp here as jnp.where doesn't work under jit
    # even if the args are static
    ib, iw = np.where((Bm[:, None] == -wm) & (Bn[:, None] == wn) & (wm != 0))
    jb, jw = np.where(
        (Bm[:, None] == wm) & (Bn[:, None] == -wn) & (wm == 0) & (wn != 0)
    )
    w_mn = put(w_mn, iw, sign(wn[iw]) * data["B_theta_mn"][ib] / jnp.abs(wm[iw]))
    w_mn = put(w_mn, jw, sign(wm[jw]) * data["B_zeta_mn"][jb] / jnp.abs(NFP * wn[jw]))
    data["w_Boozer_mn"] = w_mn
    return data


@register_compute_fun(
    name="w_Boozer",
    label="w_{\\mathrm{Boozer}}",
    units="T \\cdot m",
    units_long="Tesla * meters",
    description="Inverse Fourier transform of RHS of eq 10 in Hirshman 1995 "
    + "'Transformation from VMEC to Boozer Coordinates'",
    dim=1,
    params=[],
    transforms={"w": [[0, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=["w_Boozer_mn"],
)
def _w(params, transforms, profiles, data, **kwargs):
    data["w_Boozer"] = transforms["w"].transform(data["w_Boozer_mn"])
    return data


@register_compute_fun(
    name="w_Boozer_t",
    label="\\partial_{\\theta} w_{\\mathrm{Boozer}}",
    units="T \\cdot m",
    units_long="Tesla * meters",
    description="Inverse Fourier transform of RHS of eq 10 in Hirshman 1995 "
    + "'Transformation from VMEC to Boozer Coordinates', poloidal derivative",
    dim=1,
    params=[],
    transforms={"w": [[0, 1, 0]]},
    profiles=[],
    coordinates="rtz",
    data=["w_Boozer_mn"],
)
def _w_t(params, transforms, profiles, data, **kwargs):
    data["w_Boozer_t"] = transforms["w"].transform(data["w_Boozer_mn"], dt=1)
    return data


@register_compute_fun(
    name="w_Boozer_z",
    label="\\partial_{\\zeta} w_{\\mathrm{Boozer}}",
    units="T \\cdot m",
    units_long="Tesla * meters",
    description="Inverse Fourier transform of RHS of eq 10 in Hirshman 1995 "
    + "'Transformation from VMEC to Boozer Coordinates', toroidal derivative",
    dim=1,
    params=[],
    transforms={"w": [[0, 0, 1]]},
    profiles=[],
    coordinates="rtz",
    data=["w_Boozer_mn"],
)
def _w_z(params, transforms, profiles, data, **kwargs):
    data["w_Boozer_z"] = transforms["w"].transform(data["w_Boozer_mn"], dz=1)
    return data


@register_compute_fun(
    name="nu",
    label="\\nu = \\zeta_{B} - \\zeta",
    units="rad",
    units_long="radians",
    description="Boozer toroidal stream function",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["w_Boozer", "G", "I", "iota", "lambda"],
)
def _nu(params, transforms, profiles, data, **kwargs):
    GI = data["G"] + data["iota"] * data["I"]
    data["nu"] = (data["w_Boozer"] - data["I"] * data["lambda"]) / GI
    return data


@register_compute_fun(
    name="nu_t",
    label="\\partial_{\\theta} \\nu",
    units="rad",
    units_long="radians",
    description="Boozer toroidal stream function, derivative wrt poloidal angle",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["w_Boozer_t", "G", "I", "iota", "lambda_t"],
)
def _nu_t(params, transforms, profiles, data, **kwargs):
    GI = data["G"] + data["iota"] * data["I"]
    data["nu_t"] = (data["w_Boozer_t"] - data["I"] * data["lambda_t"]) / GI
    return data


@register_compute_fun(
    name="nu_z",
    label="\\partial_{\\zeta} \\nu",
    units="rad",
    units_long="radians",
    description="Boozer toroidal stream function, derivative wrt toroidal angle",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["w_Boozer_z", "G", "I", "iota", "lambda_z"],
)
def _nu_z(params, transforms, profiles, data, **kwargs):
    GI = data["G"] + data["iota"] * data["I"]
    data["nu_z"] = (data["w_Boozer_z"] - data["I"] * data["lambda_z"]) / GI
    return data


@register_compute_fun(
    name="theta_B",
    label="\\theta_{B}",
    units="rad",
    units_long="radians",
    description="Boozer poloidal angular coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["theta", "lambda", "iota", "nu"],
)
def _theta_B(params, transforms, profiles, data, **kwargs):
    data["theta_B"] = data["theta"] + data["lambda"] + data["iota"] * data["nu"]
    return data


@register_compute_fun(
    name="zeta_B",
    label="\\zeta_{B}",
    units="rad",
    units_long="radians",
    description="Boozer toroidal angular coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["zeta", "nu"],
)
def _zeta_B(params, transforms, profiles, data, **kwargs):
    data["zeta_B"] = data["zeta"] + data["nu"]
    return data


@register_compute_fun(
    name="sqrt(g)_B",
    label="\\sqrt{g}_{B}",
    units="~",
    units_long="None",
    description="Jacobian determinant of Boozer coordinates",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["lambda_t", "lambda_z", "nu_t", "nu_z", "iota"],
)
def _sqrtg_B(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_B"] = (1 + data["lambda_t"]) * (1 + data["nu_z"]) + (
        data["iota"] - data["lambda_z"]
    ) * data["nu_t"]
    return data


@register_compute_fun(
    name="|B|_mn",
    label="B_{mn}^{\\mathrm{Boozer}}",
    units="T",
    units_long="Tesla",
    description="Boozer harmonics of magnetic field",
    dim=1,
    params=[],
    transforms={"B": [[0, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=["sqrt(g)_B", "|B|", "rho", "theta_B", "zeta_B"],
)
def _B_mn(params, transforms, profiles, data, **kwargs):
    nodes = jnp.array([data["rho"], data["theta_B"], data["zeta_B"]]).T
    norm = 2 ** (3 - jnp.sum((transforms["B"].basis.modes == 0), axis=1))
    data["|B|_mn"] = (
        norm  # 1 if m=n=0, 2 if m=0 or n=0, 4 if m!=0 and n!=0
        * (transforms["B"].basis.evaluate(nodes).T @ (data["sqrt(g)_B"] * data["|B|"]))
        / transforms["B"].grid.num_nodes
    )
    return data


@register_compute_fun(
    name="B modes",
    label="\\mathrm{Boozer~modes}",
    units="~",
    units_long="None",
    description="Boozer harmonics",
    dim=1,
    params=[],
    transforms={"B": [[0, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _B_modes(params, transforms, profiles, data, **kwargs):
    data["B modes"] = transforms["B"].basis.modes
    return data


@register_compute_fun(
    name="f_C",
    label="(M \\iota - N) (\\mathbf{B} \\times \\nabla \\psi) \\cdot \\nabla B"
    + " - (M G + N I) \\mathbf{B} \\cdot \\nabla B",
    units="T^{3}",
    units_long="Tesla cubed",
    description="Two-term quasisymmetry metric",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["iota", "B0", "B_theta", "B_zeta", "|B|_t", "|B|_z", "G", "I", "B*grad(|B|)"],
    helicity="helicity",
)
def _f_C(params, transforms, profiles, data, **kwargs):
    M, N = kwargs.get("helicity", (1, 0))
    data["f_C"] = (M * data["iota"] - N) * data["B0"] * (
        data["B_zeta"] * data["|B|_t"] - data["B_theta"] * data["|B|_z"]
    ) - (M * data["G"] + N * data["I"]) * data["B*grad(|B|)"]
    return data


@register_compute_fun(
    name="f_T",
    label="\\nabla \\psi \\times \\nabla B \\cdot \\nabla "
    + "(\\mathbf{B} \\cdot \\nabla B)",
    units="T^{4} \\cdot m^{-2}",
    units_long="Tesla quarted / square meters",
    description="Triple product quasisymmetry metric",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["B0", "|B|_t", "|B|_z", "(B*grad(|B|))_t", "(B*grad(|B|))_z"],
)
def _f_T(params, transforms, profiles, data, **kwargs):
    data["f_T"] = data["B0"] * (
        data["|B|_t"] * data["(B*grad(|B|))_z"]
        - data["|B|_z"] * data["(B*grad(|B|))_t"]
    )
    return data


@register_compute_fun(
    name="isodynamicity",
    label="1/|B|^2 (\\mathbf{b} \\times \\nabla B) \\cdot \\nabla \\psi",
    units="~",
    units_long="None",
    description="Measure of cross field drift at each point, "
    + "unweighted by particle energy",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["b", "grad(|B|)", "|B|", "grad(psi)"],
)
def _isodynamicity(params, transforms, profiles, data, **kwargs):
    data["isodynamicity"] = (
        dot(cross(data["b"], data["grad(|B|)"]), data["grad(psi)"]) / data["|B|"] ** 2
    )
    return data



"""Profile objects for representing pressure, rotational transform, etc."""

import warnings
from abc import ABC, abstractmethod

import numpy as np
import scipy.optimize

from desc.backend import jit, jnp, put, sign
from desc.basis import FourierZernikeBasis, PowerSeries, polyder_vec, polyval_vec
from desc.derivatives import Derivative
from desc.grid import Grid, _Grid
from desc.interpolate import interp1d
from desc.io import IOAble
from desc.utils import combination_permutation, copy_coeffs, multinomial_coefficients


class _Profile(IOAble, ABC):
    """Abstract base class for profiles.

    All profile classes inherit from this, and must implement
    the compute() methods.

    The compute method should take an array of nodes and an optional array of parameters
    and compute the value or derivative of the profile at the specified nodes.
    If the parameters are not given, the ones assigned to the profile should be used.

    Subclasses must also implement getter and setter methods for params
    """

    _io_attrs_ = ["_name", "_params"]

    def __init__(self, name=""):
        self.name = name

    @property
    def name(self):
        """str: Name of the profile."""
        return self._name

    @name.setter
    def name(self, new):
        self._name = new

    @property
    @abstractmethod
    def params(self):
        """ndarray: Parameters for computation."""

    @params.setter
    @abstractmethod
    def params(self, new):
        """Set default params for computation."""

    @abstractmethod
    def compute(self, grid, params=None, dr=0, dt=0, dz=0):
        """Compute values on specified nodes, default to using self.params."""

    def to_powerseries(self, order=6, xs=100, sym="auto", rcond=None, w=None):
        """Convert this profile to a PowerSeriesProfile.

        Parameters
        ----------
        order : int
            polynomial order
        xs : int or ndarray
            x locations to use for fit. If an integer, uses that many points linearly
            spaced between 0,1
        sym : bool or "auto"
            Whether to enforce explicit even parity
        rcond : float
            Relative condition number of the fit. Singular values smaller than this
            relative to the largest singular value will be ignored. The default value
            is len(x)*eps, where eps is the relative precision of the float type, about
            2e-16 in most cases.
        w : array-like, shape(M,)
            Weights to apply to the y-coordinates of the sample points. For gaussian
            uncertainties, use 1/sigma (not 1/sigma**2).

        Returns
        -------
        profile : PowerSeriesProfile
            profile in power series form.

        """
        if jnp.isscalar(xs):
            xs = jnp.linspace(0, 1, xs)
        fs = self(xs)
        p = PowerSeriesProfile.from_values(xs, fs, order, rcond=rcond, w=w, sym=sym)
        p.name = self.name
        return p

    def to_fourierzernike(self, L=6, M=0, N=0, NFP=1, xs=100, w=None):
        """Convert this profile to a FourierZernikeProfile.

        Parameters
        ----------
        L, M, N : int
           maximum mode numbers
        NFP : int
            number of field periods
        xs : int or ndarray
            x locations to use for fit. If an integer, uses that many points linearly
            spaced between 0,1
        w : array-like, shape(M,)
            Weights to apply to the y-coordinates of the sample points. For gaussian
            uncertainties, use 1/sigma (not 1/sigma**2).

        Returns
        -------
        profile : FourierZernikeProfile
            profile in power series form.

        """
        if jnp.isscalar(xs):
            xs = jnp.linspace(0, 1, xs)
        f = self(xs)
        r = xs
        t = jnp.zeros_like(xs)
        z = jnp.zeros_like(xs)
        p = FourierZernikeProfile.from_values(r, t, z, f, L, M, N, NFP, w, self.name)
        return p

    def to_spline(self, knots=20, method="cubic2"):
        """Convert this profile to a SplineProfile.

        Parameters
        ----------
        knots : int or ndarray
            x locations to use for spline. If an integer, uses that many points linearly
            spaced between 0,1
        method : str
            method of interpolation
            - `'nearest'`: nearest neighbor interpolation
            - `'linear'`: linear interpolation
            - `'cubic'`: C1 cubic splines (aka local splines)
            - `'cubic2'`: C2 cubic splines (aka natural splines)
            - `'catmull-rom'`: C1 cubic centripetal "tension" splines

        Returns
        -------
        profile : SplineProfile
            profile in spline form.

        """
        if jnp.isscalar(knots):
            knots = jnp.linspace(0, 1, knots)
        values = self(knots)
        return SplineProfile(values, knots, method, self.name)

    def to_mtanh(
        self, order=4, xs=100, w=None, p0=None, pmax=None, pmin=None, **kwargs
    ):
        """Convert this profile to modified hyperbolic tangent + poly form.

        Parameters
        ----------
        order : int
            order of the core polynomial to fit
        xs : int or array-like, shape(M,)
            coordinate locations to evaluate for fitting. If an integer, assumes
            that many linearly spaced ints in (0,1)
        w : array-like, shape(M,)
            Weights to apply to the y-coordinates of the sample points. For gaussian
            uncertainties, use 1/sigma (not 1/sigma**2).
        p0 : array-like, shape(5+order,)
            initial guess for parameter values
        pmin : float or array-like, shape(5+order,)
            lower bounds for parameter values
        pmax : float or array-like, shape(5+order,)
            upper bounds for parameter values

        Returns
        -------
        profile : MTanhProfile
            profile in mtanh + polynomial form.

        """
        if jnp.isscalar(xs):
            xs = jnp.linspace(0, 1, xs)
        ys = self(xs)
        return MTanhProfile.from_values(
            xs,
            ys,
            order=order,
            w=w,
            p0=p0,
            pmax=pmax,
            pmin=pmin,
            name=self.name,
            **kwargs,
        )

    def __call__(self, grid, params=None, dr=0, dt=0, dz=0):
        """Evaluate the profile at a given set of points."""
        if not isinstance(grid, _Grid):
            grid = jnp.atleast_1d(grid)
            if grid.ndim == 1:
                grid = jnp.array([grid, jnp.zeros_like(grid), jnp.zeros_like(grid)]).T
            grid = Grid(grid, sort=False)
        return self.compute(grid, params, dr, dt, dz)

    def __repr__(self):
        """Get the string form of the object."""
        return (
            type(self).__name__
            + " at "
            + str(hex(id(self)))
            + " (name={})".format(self.name)
        )

    def __mul__(self, x):
        """Multiply this profile by another or a constant."""
        if np.isscalar(x):
            return ScaledProfile(x, self)
        elif isinstance(x, _Profile):
            return ProductProfile(self, x)
        else:
            raise NotImplementedError()

    def __rmul__(self, x):
        """Multiply this profile by another or a constant."""
        return self.__mul__(x)

    def __add__(self, x):
        """Add this profile with another."""
        if isinstance(x, _Profile):
            return SumProfile(self, x)
        else:
            raise NotImplementedError()

    def __neg__(self):
        """Invert the sign of this profile."""
        return ScaledProfile(-1, self)

    def __sub__(self, x):
        """Subtract another profile from this one."""
        return self.__add__(-x)


class ScaledProfile(_Profile):
    """Profile times a constant value.

    f_1(x) = a*f(x)

    Parameters
    ----------
    profile : Profile
        Base profile to scale.
    scale : float
        Scale factor.

    """

    _io_attrs_ = _Profile._io_attrs_ + ["_profile", "_scale"]

    def __init__(self, scale, profile, **kwargs):
        assert isinstance(
            profile, _Profile
        ), "profile in a ScaledProfile must be a Profile or subclass, got {}.".format(
            str(profile)
        )
        assert np.isscalar(scale), "scale must be a scalar."

        self._profile = profile.copy()
        self._scale = scale

        kwargs.setdefault("name", profile.name)
        super().__init__(**kwargs)

    @property
    def params(self):
        """ndarray: Parameters for computation [scale, profile.params]."""
        return jnp.concatenate([jnp.atleast_1d(self._scale), self._profile.params])

    @params.setter
    def params(self, x):
        self._scale, self._profile.params = self._parse_params(x)

    def _parse_params(self, x):
        if x is None:
            scale = self._scale
            params = self._profile.params
        elif isinstance(x, (tuple, list)) and len(x) == 2:
            params = x[1]
            scale = x[0]
        elif np.isscalar(x):
            scale = x
            params = self._profile.params
        elif len(x) == len(self._profile.params):
            scale = self._scale
            params = x
        elif len(x) == len(self.params):
            scale = x[0]
            params = x[1:]
        else:
            raise ValueError("Got wrong number of parameters for ScaledProfile")
        return scale, params

    def compute(self, grid, params=None, dr=0, dt=0, dz=0):
        """Compute values of profile at specified nodes.

        Parameters
        ----------
        grid : Grid
            locations to compute values at.
        params : array-like
            Parameters to use. If not given, uses the
            values given by the self.params attribute.
        dr, dt, dz : int
            derivative order in rho, theta, zeta.

        Returns
        -------
        values : ndarray
            values of the profile or its derivative at the points specified.

        """
        scale, params = self._parse_params(params)
        f = self._profile.compute(grid, params, dr, dt, dz)
        return scale * f

    def __repr__(self):
        """Get the string form of the object."""
        s = super().__repr__()
        s = s[:-1]
        s += ", scale={})".format(self._scale)
        return s


class SumProfile(_Profile):
    """Sum of two or more Profiles.

    f(x) = f1(x) + f2(x) + f3(x) ...

    Parameters
    ----------
    profiles : Profile
        Profiles to sum.

    """

    _io_attrs_ = _Profile._io_attrs_ + ["_profiles"]

    def __init__(self, *profiles, **kwargs):
        self._profiles = []
        for profile in profiles:
            assert isinstance(profile, _Profile), (
                "Each profile in a SumProfile must be a Profile or "
                + "subclass, got {}.".format(str(profile))
            )
            if isinstance(profile, SumProfile):
                self._profiles += [pro.copy() for pro in profile._profiles]
            else:
                self._profiles.append(profile.copy())
        super().__init__(**kwargs)

    @property
    def params(self):
        """ndarray: Concatenated array of parameters for computation."""
        return jnp.concatenate([profile.params for profile in self._profiles])

    @params.setter
    def params(self, x):
        x = self._parse_params(x)
        for i, profile in enumerate(self._profiles):
            profile.params = x[i]

    def _parse_params(self, x):
        if x is None:
            params = [profile.params for profile in self._profiles]
        elif isinstance(x, (list, tuple)) and len(x) == len(self._profiles):
            params = x
        elif len(x) == len(self.params):
            params = []
            i = 0
            for profile in self._profiles:
                k = len(profile.params)
                params += [x[i : i + k]]
                i += k
        else:
            raise ValueError("Got wrong number of parameters for SumProfile")
        return params

    def compute(self, grid, params=None, dr=0, dt=0, dz=0):
        """Compute values of profile at specified nodes.

        Parameters
        ----------
        grid : Grid
            locations to compute values at.
        params : array-like
            Parameters to use. If not given, uses the
            values given by the self.params attribute.
        dr, dt, dz : int
            derivative order in rho, theta, zeta.

        Returns
        -------
        values : ndarray
            values of the profile or its derivative at the points specified.

        """
        params = self._parse_params(params)
        f = 0
        for i, profile in enumerate(self._profiles):
            f += profile.compute(grid, params[i], dr, dt, dz)
        return f

    def __repr__(self):
        """Get the string form of the object."""
        s = super().__repr__()
        s = s[:-1]
        s += ", with {} profiles)".format(len(self._profiles))
        return s


class ProductProfile(_Profile):
    """Product of two or more Profiles.

    f(x) = f1(x) * f2(x) * f3(x) ...

    Parameters
    ----------
    profiles : Profile
        Profiles to multiply.

    """

    _io_attrs_ = _Profile._io_attrs_ + ["_profiles"]

    def __init__(self, *profiles, **kwargs):
        self._profiles = []
        for profile in profiles:
            assert isinstance(profile, _Profile), (
                "Each profile in a ProductProfile must be a Profile or "
                + "subclass, got {}.".format(str(profile))
            )
            if isinstance(profile, ProductProfile):
                self._profiles += [pro.copy() for pro in profile._profiles]
            else:
                self._profiles.append(profile.copy())
        super().__init__(**kwargs)

    @property
    def params(self):
        """ndarray: Concatenated array of parameters for computation."""
        return jnp.concatenate([profile.params for profile in self._profiles])

    @params.setter
    def params(self, x):
        x = self._parse_params(x)
        for i, profile in enumerate(self._profiles):
            profile.params = x[i]

    def _parse_params(self, x):
        if x is None:
            params = [profile.params for profile in self._profiles]
        elif isinstance(x, (list, tuple)) and len(x) == len(self._profiles):
            params = x
        elif len(x) == len(self.params):
            params = []
            i = 0
            for profile in self._profiles:
                k = len(profile.params)
                params += [x[i : i + k]]
                i += k
        else:
            raise ValueError("Got wrong number of parameters for ProductProfile")
        return params

    def compute(self, grid, params=None, dr=0, dt=0, dz=0):
        """Compute values of profile at specified nodes.

        Parameters
        ----------
        grid : Grid
            locations to compute values at.
        params : array-like
            Parameters to use. If not given, uses the
            values given by the self.params attribute.
        dr, dt, dz : int
            derivative order in rho, theta, zeta.

        Returns
        -------
        values : ndarray
            values of the profile or its derivative at the points specified.

        """
        if dt > 0 or dz > 0:
            raise NotImplementedError(
                "Poloidal and toroidal derivatives of ProductProfiles have not "
                + "been implemented yet"
            )
        params = self._parse_params(params)
        f = 0
        derivs = combination_permutation(len(self._profiles), dr)
        coeffs = multinomial_coefficients(len(self._profiles), dr)
        for j, drj in enumerate(derivs):
            fi = 1
            for i, profile in enumerate(self._profiles):
                fi *= profile.compute(grid, params[i], drj[i], 0, 0)
            f += coeffs[j] * fi
        return f

    def __repr__(self):
        """Get the string form of the object."""
        s = super().__repr__()
        s = s[:-1]
        s += ", with {} profiles)".format(len(self._profiles))
        return s


class PowerSeriesProfile(_Profile):
    """Profile represented by a monic power series.

    f(x) = a[0] + a[1]*x + a[2]*x**2 + ...

    Parameters
    ----------
    params: array-like
        Coefficients of the series. Assumed to be zero if not specified.
        If modes is not supplied, assumed to be in ascending  order with no
        missing values. If modes is given, coefficients can be in any order or
        indexing.
    modes : array-like
        Mode numbers for the associated coefficients. eg a[modes[i]] = params[i]
    sym : bool
        Whether the basis should only contain even powers (True) or all powers (False).
    name : str
        Name of the profile.
    """

    _io_attrs_ = _Profile._io_attrs_ + ["_basis"]

    def __init__(self, params=None, modes=None, sym="auto", name=""):
        super().__init__(name)

        if params is None:
            params = [0]
        params = np.atleast_1d(params)

        if sym == "auto":  # sym = "even" if all odd modes are zero, else sym = False
            if modes is None:
                modes = np.arange(params.size)
            else:
                modes = np.atleast_1d(modes)
            sym = np.all(params[modes % 2 != 0] == 0)
        sym = "even" if sym else False
        if modes is None:
            if sym:
                modes = np.arange(2 * params.size, step=2)
            else:
                modes = np.arange(params.size)
        else:
            modes = np.atleast_1d(modes)
        self._basis = PowerSeries(L=int(np.max(abs(modes))), sym=sym)
        self._params = np.zeros(self.basis.num_modes, dtype=float)
        for m, c in zip(modes, params):
            idx = np.where(self.basis.modes[:, 0] == int(m))[0]
            self._params[idx] = c

    def __repr__(self):
        """Get the string form of the object."""
        s = super().__repr__()
        s = s[:-1]
        s += ", basis={})".format(self.basis)
        return s

    @property
    def sym(self):
        """str: Symmetry type of the power series."""
        return self.basis.sym

    @property
    def basis(self):
        """PowerSeriesBasis: Spectral basis for power series."""
        return self._basis

    @property
    def params(self):
        """ndarray: Parameter values."""
        return self._params

    @params.setter
    def params(self, new):
        new = jnp.atleast_1d(new)
        if new.size == self._basis.num_modes:
            self._params = jnp.asarray(new)
        else:
            raise ValueError(
                "params should have the same size as the basis, "
                + f"got {len(new)} for basis with {self._basis.num_modes} modes"
            )

    def get_params(self, l):
        """Get power series coefficients for given mode number(s)."""
        l = np.atleast_1d(l).astype(int)
        a = np.zeros_like(l).astype(float)

        idx = np.where(l[:, np.newaxis] == self.basis.modes[:, 0])

        a[idx[0]] = self.params[idx[1]]
        return a

    def set_params(self, l, a=None):
        """Set specific power series coefficients."""
        l, a = np.atleast_1d(l), np.atleast_1d(a)
        a = np.broadcast_to(a, l.shape)
        for ll, aa in zip(l, a):
            idx = self.basis.get_idx(ll, 0, 0)
            if aa is not None:
                self.params = put(self.params, idx, aa)

    def change_resolution(self, L, M=None, N=None):
        """Set a new maximum mode number."""
        modes_old = self.basis.modes
        self.basis.change_resolution(L)
        self.params = copy_coeffs(self.params, modes_old, self.basis.modes)

    def compute(self, grid, params=None, dr=0, dt=0, dz=0):
        """Compute values of profile at specified nodes.

        Parameters
        ----------
        grid : Grid
            locations to compute values at.
        params : array-like
            polynomial coefficients to use, in ascending order. If not given, uses the
            values given by the params attribute
        dr, dt, dz : int
            derivative order in rho, theta, zeta

        Returns
        -------
        values : ndarray
            values of the profile or its derivative at the points specified

        """
        if params is None:
            params = self.params
        if (dt != 0) or (dz != 0):
            return jnp.zeros(grid.num_nodes)
        if self.sym:
            # need to pad with odd numbered modes
            params = jnp.array([params, jnp.zeros_like(params)]).flatten(order="F")
        r = grid.nodes[grid.unique_rho_idx, 0]
        f = polyval_vec(polyder_vec(jnp.atleast_2d(params[::-1]), dr, False), r)[0]
        return f[grid.inverse_rho_idx]

    @classmethod
    def from_values(cls, x, y, order=6, rcond=None, w=None, sym="auto", name=""):
        """Fit a PowerSeriesProfile from point data.

        Parameters
        ----------
        x : array-like, shape(M,)
            coordinate locations
        y : array-like, shape(M,)
            function values
        order : int
            order of the polynomial to fit
        rcond : float
            Relative condition number of the fit. Singular values smaller than this
            relative to the largest singular value will be ignored. The default value
            is len(x)*eps, where eps is the relative precision of the float type, about
            2e-16 in most cases.
        w : array-like, shape(M,)
            Weights to apply to the y-coordinates of the sample points. For gaussian
            uncertainties, use 1/sigma (not 1/sigma**2).
        sym : bool
            Whether the basis should only contain even powers (T) or all powers (F).
        name : str
            name of the profile

        Returns
        -------
        profile : PowerSeriesProfile
            profile in power series basis fit to given data.

        """
        if sym and sym != "auto":
            x = x**2
            order = order // 2
        params = jnp.polyfit(x, y, order, rcond=rcond, w=w, full=False)[::-1]
        return cls(params, sym=sym, name=name)


class SplineProfile(_Profile):
    """Profile represented by a piecewise cubic spline.

    Parameters
    ----------
    params: array-like
        Values of the function at knot locations.
    knots : int or ndarray
        x locations to use for spline. If an integer, uses that many points linearly
        spaced between 0,1
    method : str
        method of interpolation
        - `'nearest'`: nearest neighbor interpolation
        - `'linear'`: linear interpolation
        - `'cubic'`: C1 cubic splines (aka local splines)
        - `'cubic2'`: C2 cubic splines (aka natural splines)
        - `'catmull-rom'`: C1 cubic centripetal "tension" splines
    name : str
        name of the profile

    """

    _io_attrs_ = _Profile._io_attrs_ + ["_knots", "_method"]

    def __init__(self, values=None, knots=None, method="cubic2", name=""):
        super().__init__(name)

        if values is None:
            values = [0, 0, 0]
        values = np.atleast_1d(values)
        if knots is None:
            knots = np.linspace(0, 1, values.size)
        else:
            knots = np.atleast_1d(knots)
        self._knots = knots
        self._params = values
        self._method = method

    def __repr__(self):
        """Get the string form of the object."""
        s = super().__repr__()
        s = s[:-1]
        s += ", method={}, num_knots={})".format(self._method, len(self._knots))
        return s

    @property
    def params(self):
        """ndarray: Parameters for computation."""
        return self._params

    @params.setter
    def params(self, new):
        if len(new) == len(self._knots):
            self._params = jnp.asarray(new)
        else:
            raise ValueError(
                "params should have the same size as the knots, "
                + f"got {len(new)} values for {len(self._knots)} knots"
            )

    def compute(self, grid, params=None, dr=0, dt=0, dz=0):
        """Compute values of profile at specified nodes.

        Parameters
        ----------
        grid : Grid
            locations to compute values at.
        params : array-like
            spline values to use. If not given, uses the
            values given by the params attribute
        dr, dt, dz : int
            derivative order in rho, theta, zeta

        Returns
        -------
        values : ndarray
            values of the profile or its derivative at the points specified

        """
        if params is None:
            params = self.params
        if dt != 0 or dz != 0:
            return jnp.zeros_like(grid.nodes[:, 0])
        xq = grid.nodes[grid.unique_rho_idx, 0]
        x = self._knots
        f = params
        fq = interp1d(xq, x, f, method=self._method, derivative=dr, extrap=True)
        return fq[grid.inverse_rho_idx]


class MTanhProfile(_Profile):
    r"""Profile represented by a modified hyperbolic tangent + polynomial.

    Profile is parameterized by pedestal height (ped, :math:`p`), SOL height
    (offset, :math:`o`), pedestal symmetry point (sym, :math:`s`), pedestal width
    (width, :math:`w`), and a polynomial:

    .. math::

        f = o + 1/2 (o - p) (\tanh(z) - 1) + 1/2 (o - p) g(y)

    Where :math:`z=(x-s)/w`, :math:`y=e^z/(e^{2z}+1)`, and :math:`g` is a polynomial
    with no constant term

    Parameters
    ----------
    params: array-like
        parameters for mtanh + poly. ``params = [ped, offset, sym, width, *core_poly]``
        where core poly are the polynomial coefficients in ascending order, without
        a constant term
    name : str
        name of the profile

    """

    def __init__(self, params=None, name=""):
        super().__init__(name)

        if params is None:
            params = [0, 0, 1, 1, 0]
        self._params = params

    def __repr__(self):
        """Get the string form of the object."""
        s = super().__repr__()
        s = s[:-1]
        s += ", num_params={})".format(len(self._params))
        return s

    @property
    def params(self):
        """ndarray: Parameter values."""
        return self._params

    @params.setter
    def params(self, new):
        new = jnp.atleast_1d(new)
        if new.size >= 5:
            self._params = jnp.asarray(new)
        else:
            raise ValueError(
                "params should have at least 5 elements [ped, offset, sym, width,"
                + f"*core_poly]  got only {new.size} values"
            )

    @staticmethod
    def _mtanh(x, ped, offset, sym, width, core_poly, dx=0):
        """Compute modified tanh + polynomial profile.

        Parameters
        ----------
        x : ndarray
            evaluation locations
        ped : float
            height of pedestal
        offset : float
            height of SOL
        sym : float
            symmetry point
        width : float
            width of pedestal
        core_poly : ndarray
            polynomial coefficients in ascending order [x^1,...x^n]
        dx : int
           radial derivative order

        Returns
        -------
        y : ndarray
            profile evaluated at x
        """
        core_poly = jnp.pad(jnp.asarray(core_poly), ((1, 0)))
        z = (x - sym) / width

        if dx == 0:
            y = 1 / 2 * (ped - offset) * (1 - jnp.tanh(z)) + offset
        elif dx == 1:
            y = -1 / (2 * width) * (1 - jnp.tanh(z) ** 2) * (ped - offset)
        elif dx == 2:
            y = (ped - offset) * (jnp.tanh(-z) ** 2 - 1) * jnp.tanh(-z) / width**2

        e2z = jnp.exp(2 * z)
        zz = z / (1 + e2z)
        if dx == 0:
            f = jnp.polyval(core_poly[::-1], zz)
        elif dx == 1:
            dz = ((1 + e2z) - 2 * z * e2z) / (width * (1 + e2z) ** 2)
            f = jnp.polyval(jnp.polyder(core_poly[::-1], 1), zz) * dz
        elif dx == 2:
            dz = ((1 + e2z) - 2 * z * e2z) / (width * (1 + e2z) ** 2)
            ddz = (
                4
                * (-width * (1 + e2z) + (1 - e2z) * (sym - x))
                * e2z
                / (width**3 * (e2z + 1) ** 3)
            )
            f = (
                jnp.polyval(jnp.polyder(core_poly[::-1], 2), zz) * dz**2
                + jnp.polyval(jnp.polyder(core_poly[::-1], 1), zz) * ddz
            )

        y = y + f * (offset - ped) / 2
        return y

    def compute(self, grid, params=None, dr=0, dt=0, dz=0):
        """Compute values of profile at specified nodes.

        Parameters
        ----------
        grid : Grid
            locations to compute values at.
        params : array-like
            coefficients to use, in order. [ped, offset, sym, width, core_poly]
            If not given, uses the values given by the params attribute
        dr, dt, dz : int
            derivative order in rho, theta, zeta

        Returns
        -------
        values : ndarray
            values of the profile or its derivative at the points specified

        """
        if params is None:
            params = self.params
        if dt != 0 or dz != 0:
            return jnp.zeros_like(grid.nodes[:, 0])

        xq = grid.nodes[grid.unique_rho_idx, 0]

        ped = params[0]
        offset = params[1]
        sym = params[2]
        width = params[3]
        core_poly = params[4:]

        y = MTanhProfile._mtanh(xq, ped, offset, sym, width, core_poly, dx=dr)
        return y[grid.inverse_rho_idx]

    @classmethod
    def from_values(
        cls,
        x,
        y,
        order=4,
        w=None,
        p0=None,
        pmax=None,
        pmin=None,
        name="",
        **kwargs,
    ):
        """Fit a MTanhProfile from point data.

        Parameters
        ----------
        x : array-like, shape(M,)
            coordinate locations
        y : array-like, shape(M,)
            function values
        order : int
            order of the core polynomial to fit
        w : array-like, shape(M,)
            Weights to apply to the y-coordinates of the sample points. For gaussian
            uncertainties, use 1/sigma (not 1/sigma**2).
        p0 : array-like, shape(4+order,)
            initial guess for parameter values [ped, offset, sym, width, core_poly].
            Use a value of "None" to use the default initial guess for that parameter
        pmin : float or array-like, shape(4+order,)
            lower bounds for parameter values
            Use a value of "None" to use the default bound for that parameter
        pmax : float or array-like, shape(4+order,)
            upper bounds for parameter values
            Use a value of "None" to use the default bound for that parameter
        name : str
            name of the profile
        kwargs :
            additional keyword arguments passed to scipy.optimize.least_squares

        Returns
        -------
        profile : MTanhProfile
            profile in mtanh + polynomial form.

        """
        if w is None:
            w = np.ones_like(x)
        fun = (
            lambda args: (
                cls._mtanh(x, args[0], args[1], args[2], args[3], args[4:]) - y
            )
            / w
        )
        jac = jit(Derivative(fun, 0, "fwd").compute)
        fun = jit(fun)
        ped0 = np.clip(interp1d([0.93], x, y, "cubic2", extrap=True), 0, np.inf)[0]
        off0 = np.clip(interp1d([0.98], x, y, "cubic2", extrap=True), 0, np.inf)[0]
        default_pmax = np.array([np.inf, np.inf, 1.02, 0.2, np.inf])
        default_pmin = np.array([0, 0, 0.9, 0.0, -np.inf])
        default_p0 = np.array([ped0, off0, 0.95, 0.1, 0])

        p0_ = np.atleast_1d(p0)
        pmin_ = np.atleast_1d(pmax)
        pmax_ = np.atleast_1d(pmin)
        p0 = np.zeros(order + 4)
        pmax = np.zeros(order + 4)
        pmin = np.zeros(order + 4)
        for i in range(order + 4):
            if i < len(p0_) and p0_[i] is not None:
                p0[i] = p0_[i]
            else:
                p0[i] = default_p0[np.clip(i, 0, len(default_p0) - 1)]
            if i < len(pmax_) and pmax_[i] is not None:
                pmax[i] = pmax_[i]
            else:
                pmax[i] = default_pmax[np.clip(i, 0, len(default_pmax) - 1)]
            if i < len(pmin_) and pmin_[i] is not None:
                pmin[i] = pmin_[i]
            else:
                pmin[i] = default_pmin[np.clip(i, 0, len(default_pmin) - 1)]

        out = scipy.optimize.least_squares(
            fun, jac=jac, x0=p0, method="trf", bounds=(pmin, pmax), **kwargs
        )
        if not out.success:
            warnings.warn("Fitting did not converge, parameters may not be correct")
        params = out.x
        return MTanhProfile(params, name)


class FourierZernikeProfile(_Profile):
    """Possibly anisotropic profile represented by Fourier-Zernike basis.

    Parameters
    ----------
    params: array-like, shape(k,)
        coefficients of the series. If modes is not supplied, assumed to be only radial
        modes in ascending order with no missing values. If modes is given, coefficients
        can be in any order or indexing.
    modes : array-like, shape(k,3)
        mode numbers for the associated coefficients. eg a[modes[i]] = params[i].
        If None, assumes params are only the m=0 n=0 modes
    sym : {"auto", "sin", "cos", False}
        Whether the basis should be stellarator symmetric.
    name : str
        name of the profile.

    """

    _io_attrs_ = _Profile._io_attrs_ + ["_basis"]

    def __init__(self, params=None, modes=None, sym="auto", NFP=1, name=""):
        super().__init__(name)

        if params is None:
            params = [0]
        params = np.atleast_1d(params)

        if modes is None:
            modes = np.hstack(
                [
                    np.atleast_2d(np.arange(0, 2 * len(params), 2)).T,
                    np.zeros((len(params), 2)),
                ]
            )
        modes = np.asarray(modes)
        assert np.all(modes.astype(int) == modes), "mode numbers should be integers"
        modes = modes.astype(int)

        L = np.max(abs(modes[:, 0]))
        M = np.max(abs(modes[:, 1]))
        N = np.max(abs(modes[:, 2]))
        if sym == "auto":
            if np.all(params[np.where(sign(modes[:, 1]) != sign(modes[:, 2]))] == 0):
                sym = "cos"
            elif np.all(params[np.where(sign(modes[:, 1]) == sign(modes[:, 2]))] == 0):
                sym = "sin"
            else:
                sym = False

        self._basis = FourierZernikeBasis(L=L, M=M, N=N, NFP=NFP, sym=sym)

        self._params = copy_coeffs(params, modes, self.basis.modes)

    def __repr__(self):
        """Get the string form of the object."""
        s = super().__repr__()
        s = s[:-1]
        s += ", basis={})".format(self.basis)
        return s

    @property
    def basis(self):
        """FourierZernikeBasis: Spectral basis for Fourier-Zernike series."""
        return self._basis

    @property
    def params(self):
        """ndarray: Parameter values."""
        return self._params

    @params.setter
    def params(self, new):
        new = jnp.atleast_1d(new)
        if new.size == self._basis.num_modes:
            self._params = jnp.asarray(new)
        else:
            raise ValueError(
                f"params should have the same size as the basis, got {new.size} "
                + f"for basis with {self._basis.num_modes} modes"
            )

    def get_params(self, l, m, n):
        """Get Fourier-Zernike coefficients for given mode number(s)."""
        l = np.atleast_1d(l).astype(int)
        m = np.atleast_1d(m).astype(int)
        n = np.atleast_1d(n).astype(int)
        a = np.zeros_like(l).astype(float)

        for i, (ll, mm, nn) in enumerate(zip(l, m, n)):
            idx = self.basis.get_idx(ll, mm, nn)
            a[i] = self.params[idx]
        return a

    def set_params(self, l, m, n, a=None):
        """Set specific Fourier-Zernike coefficients."""
        l, m, n, a = map(np.atleast_1d, (l, m, n, a))
        a = np.broadcast_to(a, l.shape)
        for ll, mm, nn, aa in zip(l, m, n, a):
            idx = self.basis.get_idx(ll, mm, nn)
            if aa is not None:
                self.params = put(self.params, idx, aa)

    def change_resolution(self, L=None, M=None, N=None):
        """Set a new maximum mode number."""
        modes_old = self.basis.modes
        L = L if L is not None else self.basis.L
        M = M if M is not None else self.basis.M
        N = N if N is not None else self.basis.N
        self.basis.change_resolution(L, M, N)
        self.params = copy_coeffs(self.params, modes_old, self.basis.modes)

    def compute(self, grid, params=None, dr=0, dt=0, dz=0):
        """Compute values of profile at specified nodes.

        Parameters
        ----------
        grid : Grid
            locations to compute values at.
        params : array-like
            Fourier-Zernike coefficients to use, in ascending order. If not given,
            uses the values given by the params attribute
        dr, dt, dz : int
            derivative order in rho, theta, zeta

        Returns
        -------
        values : ndarray
            values of the profile or its derivative at the points specified

        """
        if params is None:
            params = self.params
        A = self.basis.evaluate(grid.nodes, [dr, dt, dz])
        return A @ params

    @classmethod
    def from_values(cls, r, t, z, f, L=6, M=0, N=0, NFP=1, w=None, name=""):
        """Fit a FourierZernikeProfile from point data.

        Parameters
        ----------
        r, t, z : array-like, shape(k,)
            coordinate locations in rho, theta, zeta
        f : array-like, shape(k,)
            function values
        L, M, N : int
            maximum mode numbers to fit
        NFP : int
            number of field periods
        w : array-like, shape(k,)
            Weights to apply to the y-coordinates of the sample points. For gaussian
            uncertainties, use 1/sigma (not 1/sigma**2).
        name : str
            name of the profile

        Returns
        -------
        profile : PowerSeriesProfile
            profile in power series basis fit to given data.

        """
        nodes = jnp.vstack([r, t, z]).T
        basis = FourierZernikeBasis(L, M, N, NFP)
        A = basis.evaluate(nodes)
        if w is not None:
            A *= w[:, np.newaxis]
            f *= w
        scale = jnp.sqrt((A * A).sum(axis=0))
        scale = jnp.where(scale == 0, 1, scale)
        A /= scale
        c, resids, rank, s = jnp.linalg.lstsq(A, f, rcond=None)
        c = (c.T / scale).T  # broadcast scale coefficients
        return cls(c, modes=basis.modes, NFP=NFP, name=name)



"""Utility functions used in optimization problems."""

import copy

import numpy as np

from desc.backend import cond, jit, jnp, put
from desc.utils import Index


def inequality_to_bounds(x0, fun, grad, hess, constraint, bounds, *args):
    """Convert inequality constraints to bounds using slack variables.

    We do this by introducing slack variables s

    ie, lb < con(x) < ub --> con(x) - s == 0, lb < s < ub

    A new state vector z is defined as [x, s] and the problem
    is transformed into one that has only equality constraints
    and simple bounds on the variables z

    Parameters
    ----------
    x0 : ndarray
        Starting point for primal variables
    fun, grad, hess : callable
        Functions for computing the objective and derivatives
    constraint : scipy.optimize.NonlinearConstraint
        constraint object of both equality and inequality constraints
    bounds : tuple
        lower and upper bounds for primal variables x

    Returns
    -------
    z0 : ndarray
        Starting point for primal + slack variables
    fun, grad, hess : callable
        functions for computing objective and derivatives wrt z
    constraint : scipy.optimize.NonlinearConstraint
        constraint containing just equality constraints
    bounds : tuple
        lower and upper bounds on combined variable z
    z2xs : callable
        function for splitting combined variable z into primal
        and slack variables x and s

    """
    c0 = constraint.fun(x0, *args)
    ncon = c0.size
    bounds = tuple(jnp.broadcast_to(bi, x0.shape) for bi in bounds)
    cbounds = (constraint.lb, constraint.ub)
    cbounds = tuple(jnp.broadcast_to(bi, c0.shape) for bi in cbounds)
    lbs, ubs = cbounds
    lbx, ubx = bounds

    ineq_mask = lbs != ubs
    eq_mask = lbs == ubs
    eq_target = lbs[~ineq_mask]
    nslack = jnp.sum(ineq_mask)
    zbounds = (
        jnp.concatenate([lbx, lbs[ineq_mask]]),
        jnp.concatenate([ubx, ubs[ineq_mask]]),
    )
    s0 = c0[ineq_mask]
    s0 = jnp.clip(s0, lbs[ineq_mask], ubs[ineq_mask])
    z0 = jnp.concatenate([x0, s0])
    target = jnp.zeros(c0.size)
    target = put(target, eq_mask, eq_target)

    def z2xs(z):
        return z[: len(z) - nslack], z[len(z) - nslack :]

    def fun_wrapped(z, *args):
        x, s = z2xs(z)
        return fun(x, *args)

    if hess is None:
        # assume grad is really jac of least squares
        def grad_wrapped(z, *args):
            x, s = z2xs(z)
            g = grad(x, *args)
            return jnp.hstack([g, jnp.zeros((g.shape[0], nslack))])

    else:

        def grad_wrapped(z, *args):
            x, s = z2xs(z)
            g = grad(x, *args)
            return jnp.concatenate([g, jnp.zeros(nslack)])

    if callable(hess):

        def hess_wrapped(z, *args):
            x, s = z2xs(z)
            H = hess(x, *args)
            return jnp.pad(H, (0, nslack))

    else:  # using BFGS
        hess_wrapped = hess

    def confun_wrapped(z, *args):
        x, s = z2xs(z)
        c = constraint.fun(x, *args)
        sbig = jnp.zeros(ncon)
        sbig = put(sbig, ineq_mask, s)
        return c - sbig - target

    def conjac_wrapped(z, *args):
        x, s = z2xs(z)
        J = constraint.jac(x, *args)
        I = jnp.eye(nslack)
        Js = jnp.zeros((ncon, nslack))
        Js = put(Js, Index[ineq_mask, :], -I)
        return jnp.hstack([J, Js])

    if callable(constraint.hess):

        def conhess_wrapped(z, y, *args):
            x, s = z2xs(z)
            H = constraint.hess(x, y, *args)
            return jnp.pad(H, (0, nslack))

    else:  # using BFGS
        conhess_wrapped = constraint.hess

    if hasattr(constraint, "vjp"):

        def vjp_wrapped(y, z, *args):
            x, s = z2xs(z)
            I = jnp.eye(nslack)
            Js = jnp.zeros((ncon, nslack))
            Js = put(Js, Index[ineq_mask, :], -I)
            vjpx = constraint.vjp(y, x, *args)
            vjps = jnp.dot(y, Js)
            return jnp.concatenate([vjpx, vjps])

    else:

        def vjp_wrapped(y, z, *args):
            J = conjac_wrapped(z, *args)
            return jnp.dot(y, J)

    newcon = copy.copy(constraint)
    newcon.fun = confun_wrapped
    newcon.jac = conjac_wrapped
    newcon.hess = conhess_wrapped
    newcon.lb = target
    newcon.ub = target
    newcon.vjp = vjp_wrapped

    return z0, fun_wrapped, grad_wrapped, hess_wrapped, newcon, zbounds, z2xs


@jit
def gershgorin_bounds(H):
    """Upper and lower bounds for eigenvalues of a square matrix.

    Given a square matrix ``H`` compute upper
    and lower bounds for its eigenvalues (Gershgorin Bounds).
    Defined ref. [1].

    References
    ----------
    [1] Conn, A. R., Gould, N. I., & Toint, P. L.
           Trust region methods. 2000. Siam. pp. 19.
    """
    H_diag = jnp.diag(H)
    H_diag_abs = jnp.abs(H_diag)
    H_row_sums = jnp.sum(jnp.abs(H), axis=1)
    lb = jnp.min(H_diag + H_diag_abs - H_row_sums)
    ub = jnp.max(H_diag - H_diag_abs + H_row_sums)

    return lb, ub


@jit
def _cholmod(A, maxiter=4):
    """Modified Cholesky factorization of indefinite matrix.

    Uses Gershgorin bounds and bisection search to find the
    smallest diagonal correction alpha such that A + alpha*I is
    positive definite, and returns the cholesky factorization
    of A + alpha*I.

    Attempts to find smallest alpha to the nearest order of magnitude,
    in maxiter steps (so 2**maxiter values of alpha will be scanned over).
    Scans over values of -log(abs(lb))< log(alpha) < log(abs(lb))

    Cost is approximately maxiter times the cost of a single cholesky
    factorization.

    Parameters
    ----------
    A : ndarray
        Matrix to factorize. Should be hermitian. No checking is done.
    maxiter : int
        Maximum number of bisection search steps

    Returns
    -------
    L : ndarray
        Lower triangular cholesky factor, such that L@L.T ~ A

    """
    assert int(maxiter) == maxiter
    maxiter = int(maxiter)
    k = 2**maxiter  # number of values to try
    A = jnp.asarray(A)
    n = A.shape[0]
    eye = jnp.eye(n)
    # upper and lower bounds on eig(A)
    lb, ub = gershgorin_bounds(A)
    # upper bound on log(alpha) such that A + alpha*I > 0, ie we know alpha < ub
    # lower bound on eig(A) = upper bound on alpha, +1 in log scale to make sure
    # that it's actually greater than the maximum alpha
    ub = jnp.log10(jnp.abs(lb)) + 1
    # we know alpha > 0 because otherwise initial factorization would have succeeded
    # but we'd like to be a bit better (in log scaling). This is just a heuristic but
    # seems ok in practice
    m = jnp.mean(jnp.abs(A))
    lb = ub - 2 * abs(ub) - abs(jnp.log10(m))
    # values to try
    alphas = jnp.logspace(lb, ub, k)
    kbest = k // 2
    klow = 0
    khigh = k
    # first we try alpha = max, which we know will succeed by gershgorin bounds
    # but might be too big a correction, so then we try to reduce it while keeping
    # A + alpha*I positive definite
    Lbest = jnp.linalg.cholesky(A + alphas[k] * eye)
    for i in range(maxiter):
        L = jnp.linalg.cholesky(A + alphas[kbest] * eye)
        # check if it succeeded
        isnan = jnp.any(jnp.isnan(L))
        # adjust bounds for correction
        klow = isnan * kbest + (1 - isnan) * klow
        khigh = isnan * khigh + (1 - isnan) * kbest
        kbest = (klow + khigh) // 2
        # if it succeeded, mark it as the best so far
        Lbest = cond(isnan, lambda _: Lbest, lambda _: L, None)
    return Lbest


@jit
def chol(A):
    """Cholesky factorization of possibly indefinite matrix.

    If matrix is positive definite, returns the regular Cholesky factorization,
    otherwise performs a modified factorization to ensure the factors are positive
    definite.

    Parameters
    ----------
    A : ndarray
        Matrix to factorize. Should be hermitian. No checking is done.

    Returns
    -------
    L : ndarray
        Lower triangular (approximate) cholesky factor, such that L@L.T ~ A

    """
    L = jnp.linalg.cholesky(A)
    L = cond(jnp.any(jnp.isnan(L)), lambda A: _cholmod(A), lambda A: L, A)
    return L


@jit
def evaluate_quadratic_form_hess(H, g, x, diag=None):
    """Compute values of a quadratic function arising in trust region subproblem.

    The function is 0.5 * x.T * (H + diag) * x + g.T * x.

    Parameters
    ----------
    H : ndarray
        Hessian matrix
    g : ndarray, shape(n,)
        Gradient, defines the linear term.
    x : ndarray, shape(n,)
        position where to evaluate quadratic form
    diag : ndarray, shape (n,), optional
        Addition diagonal part, affects the quadratic term.
        If None, assumed to be 0.

    Returns
    -------
    values : float
        Value of the function.
    """
    q = x @ H @ x
    if diag is not None:
        q += jnp.sum(diag * x**2, axis=-1)
    l = jnp.dot(g, x)

    return l + 1 / 2 * q


@jit
def evaluate_quadratic_form_jac(J, g, s, diag=None):
    """Compute values of a quadratic function arising in least squares.

    The function is 0.5 * s.T * (J.T * J + diag) * s + g.T * s.

    Parameters
    ----------
    J : ndarray, sparse matrix or LinearOperator, shape (m, n)
        Jacobian matrix, affects the quadratic term.
    g : ndarray, shape (n,)
        Gradient, defines the linear term.
    s : ndarray, shape (k, n) or (n,)
        Array containing steps as rows.
    diag : ndarray, shape (n,), optional
        Addition diagonal part, affects the quadratic term.
        If None, assumed to be 0.

    Returns
    -------
    values : ndarray with shape (k,) or float
        Values of the function. If `s` was 2-D, then ndarray is
        returned, otherwise, float is returned.
    """
    if s.ndim == 1:
        Js = J.dot(s)
        q = jnp.dot(Js, Js)
        if diag is not None:
            q += jnp.dot(s * diag, s)
    else:
        Js = J.dot(s.T)
        q = jnp.sum(Js**2, axis=0)
        if diag is not None:
            q += jnp.sum(diag * s**2, axis=-1)

    l = jnp.dot(s, g)

    return 0.5 * q + l


def print_header_nonlinear(constrained=False, *args):
    """Print a pretty header."""
    s = "{:^15}{:^15}{:^15}{:^15}{:^15}{:^15}".format(
        "Iteration",
        "Total nfev",
        "Cost",
        "Cost reduction",
        "Step norm",
        "Optimality",
    )
    if constrained:
        s += "{:^15}".format("Constr viol.")
    for arg in args:
        s += "{:^15}".format(arg)
    print(s)


def print_iteration_nonlinear(
    iteration,
    nfev,
    cost,
    cost_reduction,
    step_norm,
    optimality,
    constr_violation=None,
    *args,
):
    """Print a line of optimizer output."""
    if iteration is None or abs(iteration) == np.inf:
        iteration = " " * 15
    else:
        iteration = "{:^15}".format(iteration)

    if nfev is None or abs(nfev) == np.inf:
        nfev = " " * 15
    else:
        nfev = "{:^15}".format(nfev)

    if cost is None or abs(cost) == np.inf:
        cost = " " * 15
    else:
        cost = "{: ^15.3e}".format(cost)

    if cost_reduction is None or abs(cost_reduction) == np.inf:
        cost_reduction = " " * 15
    else:
        cost_reduction = "{: ^15.3e}".format(cost_reduction)

    if step_norm is None or abs(step_norm) == np.inf:
        step_norm = " " * 15
    else:
        step_norm = "{:^15.3e}".format(step_norm)

    if optimality is None or abs(optimality) == np.inf:
        optimality = " " * 15
    else:
        optimality = "{:^15.3e}".format(optimality)
    s = "{}{}{}{}{}{}".format(
        iteration, nfev, cost, cost_reduction, step_norm, optimality
    )
    if constr_violation is not None:
        s += "{:^15.3e}".format(constr_violation)
    for arg in args:
        s += "{:^15.3e}".format(arg)
    print(s)


STATUS_MESSAGES = {
    "success": "Optimization terminated successfully.",
    "xtol": "`xtol` condition satisfied.",
    "ftol": "`ftol` condition satisfied.",
    "gtol": "`gtol` condition satisfied.",
    "max_nfev": "Maximum number of function evaluations has been exceeded.",
    "maxiter": "Maximum number of iterations has been exceeded.",
    "pr_loss": "Desired error not necessarily achieved due to precision loss.",
    "nan": "NaN result encountered.",
    "out_of_bounds": "The result is outside of the provided bounds.",
    "err": "A linalg error occurred, such as a non-psd Hessian.",
    "approx": "A bad approximation caused failure to predict improvement.",
    "callback": "User supplied callback triggered termination",
    None: None,
}


def check_termination(
    dF,
    F,
    dx_norm,
    x_norm,
    g_norm,
    reduction_ratio,
    ftol,
    xtol,
    gtol,
    iteration,
    maxiter,
    nfev,
    max_nfev,
    **kwargs,
):
    """Check termination condition and get message."""
    ftol_satisfied = 0 < dF < abs(ftol * F) and reduction_ratio > 0.25
    xtol_satisfied = dx_norm < xtol * (xtol + x_norm) and reduction_ratio > 0.25
    gtol_satisfied = g_norm < gtol
    ctol_satisfied = kwargs.get("constr_violation", 0) < kwargs.get("ctol", np.inf)

    if ctol_satisfied and any([ftol_satisfied, xtol_satisfied, gtol_satisfied]):
        message = STATUS_MESSAGES["success"]
        success = True
        if ftol_satisfied:
            message += "\n" + STATUS_MESSAGES["ftol"]
        if xtol_satisfied:
            message += "\n" + STATUS_MESSAGES["xtol"]
        if gtol_satisfied:
            message += "\n" + STATUS_MESSAGES["gtol"]
    elif iteration >= maxiter:
        success = False
        message = STATUS_MESSAGES["maxiter"]
    elif nfev >= max_nfev:
        success = False
        message = STATUS_MESSAGES["max_nfev"]
    elif dx_norm < kwargs.get("min_trust_radius", np.finfo(x_norm.dtype).eps):
        success = False
        message = STATUS_MESSAGES["approx"]
    elif kwargs.get("dx_total", 0) > kwargs.get("max_dx", np.inf):
        success = False
        message = STATUS_MESSAGES["out_of_bounds"]
    else:
        success = None
        message = None

    return success, message


@jit
def compute_jac_scale(A, prev_scale_inv=None):
    """Compute scaling factor based on column norm of Jacobian matrix."""
    scale_inv = jnp.sum(A**2, axis=0) ** 0.5
    scale_inv = jnp.where(
        scale_inv < jnp.finfo(A.dtype).eps * max(A.shape), 1, scale_inv
    )

    if prev_scale_inv is not None:
        scale_inv = jnp.maximum(scale_inv, prev_scale_inv)
    return 1 / scale_inv, scale_inv


@jit
def compute_hess_scale(H, prev_scale_inv=None):
    """Compute scaling factors based on diagonal of Hessian matrix."""
    scale_inv = jnp.abs(jnp.diag(H))
    scale_inv = jnp.where(
        scale_inv < jnp.finfo(H.dtype).eps * max(H.shape), 1, scale_inv
    )

    if prev_scale_inv is not None:
        scale_inv = jnp.maximum(scale_inv, prev_scale_inv)
    return 1 / scale_inv, scale_inv


def f_where_x(x, xs, fs, dim=0):
    """Return fs where x==xs.

    Parameters
    ----------
    x : ndarray, shape(k,)
        array to find
    xs : list of ndarray of shape(k,)
        list to compare x against
    fs : list of float, ndarray
        list of values to return value from
    dim : int
        number of dimensions the output should have

    Returns
    -------
    f : float or ndarray
        value of fs[i] where x==xs[i]
    """
    x, xs, fs = map(np.asarray, (x, xs, fs))
    assert len(xs) == len(fs)
    assert len(xs) == 0 or x.shape == xs[0].shape
    eps = np.finfo(x.dtype).eps
    i = np.where(np.all(np.isclose(x, xs, rtol=eps, atol=eps), axis=1))[0]
    # sometimes two things are within eps of x, we want the most recent one
    if len(i) > 1:
        i = i[-1]
    f = fs[i].squeeze()
    if dim == 1:
        f = np.atleast_1d(f)
    if dim == 2:
        f = np.atleast_2d(f).reshape((-1, x.size))
    return f



"""Classes for parameterized 3D space curves."""

import numbers

import numpy as np

from desc.backend import jnp, put
from desc.basis import FourierSeries
from desc.compute import rpz2xyz
from desc.grid import LinearGrid
from desc.io import InputReader
from desc.transform import Transform
from desc.utils import copy_coeffs, errorif

from .core import Curve

__all__ = ["FourierRZCurve", "FourierXYZCurve", "FourierPlanarCurve", "SplineXYZCurve"]


class FourierRZCurve(Curve):
    """Curve parameterized by Fourier series for R,Z in terms of toroidal angle phi.

    Parameters
    ----------
    R_n, Z_n: array-like
        Fourier coefficients for R, Z.
    modes_R : array-like, optional
        Mode numbers associated with R_n. If not given defaults to [-n:n].
    modes_Z : array-like, optional
        Mode numbers associated with Z_n, If not given defaults to [-n:n]].
    NFP : int
        Number of field periods.
    sym : bool
        Whether to enforce stellarator symmetry.
    name : str
        Name for this curve.

    """

    _io_attrs_ = Curve._io_attrs_ + [
        "_R_n",
        "_Z_n",
        "_R_basis",
        "_Z_basis",
        "_sym",
        "_NFP",
    ]

    def __init__(
        self,
        R_n=10,
        Z_n=0,
        modes_R=None,
        modes_Z=None,
        NFP=1,
        sym="auto",
        name="",
    ):
        super().__init__(name)
        R_n, Z_n = np.atleast_1d(R_n), np.atleast_1d(Z_n)
        if modes_R is None:
            modes_R = np.arange(-(R_n.size // 2), R_n.size // 2 + 1)
        if modes_Z is None:
            modes_Z = np.arange(-(Z_n.size // 2), Z_n.size // 2 + 1)

        if R_n.size == 0:
            raise ValueError("At least 1 coefficient for R must be supplied")
        if Z_n.size == 0:
            Z_n = np.array([0.0])
            modes_Z = np.array([0])

        modes_R, modes_Z = np.asarray(modes_R), np.asarray(modes_Z)

        assert issubclass(modes_R.dtype.type, np.integer)
        assert issubclass(modes_Z.dtype.type, np.integer)

        if sym == "auto":
            if np.all(R_n[modes_R < 0] == 0) and np.all(Z_n[modes_Z >= 0] == 0):
                sym = True
            else:
                sym = False
        self._sym = sym
        NR = np.max(abs(modes_R))
        NZ = np.max(abs(modes_Z))
        N = max(NR, NZ)
        self._NFP = NFP
        self._R_basis = FourierSeries(N, NFP, sym="cos" if sym else False)
        self._Z_basis = FourierSeries(N, NFP, sym="sin" if sym else False)

        self._R_n = copy_coeffs(R_n, modes_R, self.R_basis.modes[:, 2])
        self._Z_n = copy_coeffs(Z_n, modes_Z, self.Z_basis.modes[:, 2])

    @property
    def sym(self):
        """Whether this curve has stellarator symmetry."""
        return self._sym

    @property
    def R_basis(self):
        """Spectral basis for R_Fourier series."""
        return self._R_basis

    @property
    def Z_basis(self):
        """Spectral basis for Z_Fourier series."""
        return self._Z_basis

    @property
    def NFP(self):
        """Number of field periods."""
        return self._NFP

    @NFP.setter
    def NFP(self, new):
        assert (
            isinstance(new, numbers.Real) and int(new) == new and new > 0
        ), f"NFP should be a positive integer, got {type(new)}"
        self.change_resolution(NFP=new)

    @property
    def N(self):
        """Maximum mode number."""
        return max(self.R_basis.N, self.Z_basis.N)

    def change_resolution(self, N=None, NFP=None, sym=None):
        """Change the maximum toroidal resolution."""
        if (
            ((N is not None) and (N != self.N))
            or ((NFP is not None) and (NFP != self.NFP))
            or (sym is not None)
            and (sym != self.sym)
        ):
            self._NFP = NFP if NFP is not None else self.NFP
            self._sym = sym if sym is not None else self.sym
            N = N if N is not None else self.N
            R_modes_old = self.R_basis.modes
            Z_modes_old = self.Z_basis.modes
            self.R_basis.change_resolution(
                N=N, NFP=self.NFP, sym="cos" if self.sym else self.sym
            )
            self.Z_basis.change_resolution(
                N=N, NFP=self.NFP, sym="sin" if self.sym else self.sym
            )
            self.R_n = copy_coeffs(self.R_n, R_modes_old, self.R_basis.modes)
            self.Z_n = copy_coeffs(self.Z_n, Z_modes_old, self.Z_basis.modes)

    def get_coeffs(self, n):
        """Get Fourier coefficients for given mode number(s)."""
        n = np.atleast_1d(n).astype(int)
        R = np.zeros_like(n).astype(float)
        Z = np.zeros_like(n).astype(float)

        idxR = np.where(n[:, np.newaxis] == self.R_basis.modes[:, 2])
        idxZ = np.where(n[:, np.newaxis] == self.Z_basis.modes[:, 2])

        R[idxR[0]] = self.R_n[idxR[1]]
        Z[idxZ[0]] = self.Z_n[idxZ[1]]
        return R, Z

    def set_coeffs(self, n, R=None, Z=None):
        """Set specific Fourier coefficients."""
        n, R, Z = np.atleast_1d(n), np.atleast_1d(R), np.atleast_1d(Z)
        R = np.broadcast_to(R, n.shape)
        Z = np.broadcast_to(Z, n.shape)
        for nn, RR, ZZ in zip(n, R, Z):
            if RR is not None:
                idxR = self.R_basis.get_idx(0, 0, nn)
                self.R_n = put(self.R_n, idxR, RR)
            if ZZ is not None:
                idxZ = self.Z_basis.get_idx(0, 0, nn)
                self.Z_n = put(self.Z_n, idxZ, ZZ)

    @property
    def R_n(self):
        """Spectral coefficients for R."""
        return self._R_n

    @R_n.setter
    def R_n(self, new):
        if len(new) == self.R_basis.num_modes:
            self._R_n = jnp.asarray(new)
        else:
            raise ValueError(
                f"R_n should have the same size as the basis, got {len(new)} for "
                + f"basis with {self.R_basis.num_modes} modes."
            )

    @property
    def Z_n(self):
        """Spectral coefficients for Z."""
        return self._Z_n

    @Z_n.setter
    def Z_n(self, new):
        if len(new) == self.Z_basis.num_modes:
            self._Z_n = jnp.asarray(new)
        else:
            raise ValueError(
                f"Z_n should have the same size as the basis, got {len(new)} for "
                + f"basis with {self.Z_basis.num_modes} modes"
            )

    @classmethod
    def from_input_file(cls, path):
        """Create a axis curve from Fourier coefficients in a DESC or VMEC input file.

        Parameters
        ----------
        path : Path-like or str
            Path to DESC or VMEC input file.

        Returns
        -------
        curve : FourierRZToroidalCurve
            Axis with given Fourier coefficients.

        """
        f = open(path)
        if "&INDATA" in f.readlines()[0].upper():  # vmec input, convert to desc
            inputs = InputReader.parse_vmec_inputs(f)[-1]
        else:
            inputs = InputReader().parse_inputs(f)[-1]
        curve = FourierRZCurve(
            inputs["axis"][:, 1],
            inputs["axis"][:, 2],
            inputs["axis"][:, 0].astype(int),
            inputs["axis"][:, 0].astype(int),
            inputs["NFP"],
            inputs["sym"],
        )
        return curve


def _unclose_curve(X, Y, Z):
    if np.allclose([X[0], Y[0], Z[0]], [X[-1], Y[-1], Z[-1]], atol=1e-14):
        closedX, closedY, closedZ = X.copy(), Y.copy(), Z.copy()
        X, Y, Z = X[:-1], Y[:-1], Z[:-1]
        flag = True
    else:
        closedX, closedY, closedZ = (
            np.append(X, X[0]),
            np.append(Y, Y[0]),
            np.append(Z, Z[0]),
        )
        flag = False
    return X, Y, Z, closedX, closedY, closedZ, flag


class FourierXYZCurve(Curve):
    """Curve parameterized by Fourier series for X,Y,Z in terms of arbitrary angle s.

    Parameters
    ----------
    X_n, Y_n, Z_n: array-like
        Fourier coefficients for X, Y, Z
    modes : array-like
        mode numbers associated with X_n etc.
    name : str
        name for this curve

    """

    _io_attrs_ = Curve._io_attrs_ + [
        "_X_n",
        "_Y_n",
        "_Z_n",
        "_X_basis",
        "_Y_basis",
        "_Z_basis",
    ]

    def __init__(
        self,
        X_n=[0, 10, 2],
        Y_n=[0, 0, 0],
        Z_n=[-2, 0, 0],
        modes=None,
        name="",
    ):
        super().__init__(name)
        X_n, Y_n, Z_n = np.atleast_1d(X_n), np.atleast_1d(Y_n), np.atleast_1d(Z_n)
        if modes is None:
            modes = np.arange(-(X_n.size // 2), X_n.size // 2 + 1)
        else:
            modes = np.asarray(modes)

        assert issubclass(modes.dtype.type, np.integer)

        N = np.max(abs(modes))
        self._X_basis = FourierSeries(N, NFP=1, sym=False)
        self._Y_basis = FourierSeries(N, NFP=1, sym=False)
        self._Z_basis = FourierSeries(N, NFP=1, sym=False)
        self._X_n = copy_coeffs(X_n, modes, self.X_basis.modes[:, 2])
        self._Y_n = copy_coeffs(Y_n, modes, self.Y_basis.modes[:, 2])
        self._Z_n = copy_coeffs(Z_n, modes, self.Z_basis.modes[:, 2])

    @property
    def X_basis(self):
        """Spectral basis for X Fourier series."""
        return self._X_basis

    @property
    def Y_basis(self):
        """Spectral basis for Y Fourier series."""
        return self._Y_basis

    @property
    def Z_basis(self):
        """Spectral basis for Z Fourier series."""
        return self._Z_basis

    @property
    def N(self):
        """Maximum mode number."""
        return max(self.X_basis.N, self.Y_basis.N, self.Z_basis.N)

    def change_resolution(self, N=None):
        """Change the maximum angular resolution."""
        if (N is not None) and (N != self.N):
            Xmodes_old = self.X_basis.modes
            Ymodes_old = self.Y_basis.modes
            Zmodes_old = self.Z_basis.modes
            self.X_basis.change_resolution(N=N)
            self.Y_basis.change_resolution(N=N)
            self.Z_basis.change_resolution(N=N)
            self.X_n = copy_coeffs(self.X_n, Xmodes_old, self.X_basis.modes)
            self.Y_n = copy_coeffs(self.Y_n, Ymodes_old, self.Y_basis.modes)
            self.Z_n = copy_coeffs(self.Z_n, Zmodes_old, self.Z_basis.modes)

    def get_coeffs(self, n):
        """Get Fourier coefficients for given mode number(s)."""
        n = np.atleast_1d(n).astype(int)
        X = np.zeros_like(n).astype(float)
        Y = np.zeros_like(n).astype(float)
        Z = np.zeros_like(n).astype(float)

        Xidx = np.where(n[:, np.newaxis] == self.X_basis.modes[:, 2])
        Yidx = np.where(n[:, np.newaxis] == self.Y_basis.modes[:, 2])
        Zidx = np.where(n[:, np.newaxis] == self.Z_basis.modes[:, 2])

        X[Xidx[0]] = self.X_n[Xidx[1]]
        Y[Yidx[0]] = self.Y_n[Yidx[1]]
        Z[Zidx[0]] = self.Z_n[Zidx[1]]
        return X, Y, Z

    def set_coeffs(self, n, X=None, Y=None, Z=None):
        """Set specific Fourier coefficients."""
        n, X, Y, Z = (
            np.atleast_1d(n),
            np.atleast_1d(X),
            np.atleast_1d(Y),
            np.atleast_1d(Z),
        )
        X = np.broadcast_to(X, n.shape)
        Y = np.broadcast_to(Y, n.shape)
        Z = np.broadcast_to(Z, n.shape)
        for nn, XX in zip(n, X):
            idx = self.X_basis.get_idx(0, 0, nn)
            if XX is not None:
                self.X_n = put(self.X_n, idx, XX)

        for nn, YY in zip(n, Y):
            idx = self.Y_basis.get_idx(0, 0, nn)
            if YY is not None:
                self.Y_n = put(self.Y_n, idx, YY)

        for nn, ZZ in zip(n, Z):
            idx = self.Z_basis.get_idx(0, 0, nn)
            if ZZ is not None:
                self.Z_n = put(self.Z_n, idx, ZZ)

    @property
    def X_n(self):
        """Spectral coefficients for X."""
        return self._X_n

    @X_n.setter
    def X_n(self, new):
        if len(new) == self.X_basis.num_modes:
            self._X_n = jnp.asarray(new)
        else:
            raise ValueError(
                f"X_n should have the same size as the basis, got {len(new)} for "
                + f"basis with {self.X_basis.num_modes} modes."
            )

    @property
    def Y_n(self):
        """Spectral coefficients for Y."""
        return self._Y_n

    @Y_n.setter
    def Y_n(self, new):
        if len(new) == self.Y_basis.num_modes:
            self._Y_n = jnp.asarray(new)
        else:
            raise ValueError(
                f"Y_n should have the same size as the basis, got {len(new)} for "
                + f"basis with {self.Y_basis.num_modes} modes."
            )

    @property
    def Z_n(self):
        """Spectral coefficients for Z."""
        return self._Z_n

    @Z_n.setter
    def Z_n(self, new):
        if len(new) == self.Z_basis.num_modes:
            self._Z_n = jnp.asarray(new)
        else:
            raise ValueError(
                f"Z_n should have the same size as the basis, got {len(new)} for "
                + f"basis with {self.Z_basis.num_modes} modes."
            )

    @classmethod
    def from_values(cls, coords, N=10, s=None, basis="xyz", name=""):
        """Fit coordinates to FourierXYZCurve representation.

        Parameters
        ----------
        coords: ndarray
            coordinates to fit a FourierXYZCurve object with.
        N : int
            Fourier resolution of the new X,Y,Z representation.
            default is 10
        s : ndarray or "arclength"
            arbitrary curve parameter to use for the fitting.
            Should be monotonic, 1D array of same length as
            coords. if None, defaults linearly spaced in [0,2pi)
            Alternative, can pass "arclength" to use normalized distance between points.
        basis : {"rpz", "xyz"}
            basis for input coordinates. Defaults to "xyz"
        Returns
        -------
        curve : FourierXYZCurve
            New representation of the curve parameterized by Fourier series for X,Y,Z.

        """
        if basis == "xyz":
            coords_xyz = coords
        else:
            coords_xyz = rpz2xyz(coords, phi=coords[:, 1])
        X = coords_xyz[:, 0]
        Y = coords_xyz[:, 1]
        Z = coords_xyz[:, 2]

        X, Y, Z, closedX, closedY, closedZ, _ = _unclose_curve(X, Y, Z)

        if isinstance(s, str):
            assert s == "arclength", f"got unknown specification for s {s}"
            # find equal arclength angle-like variable, and use that as theta
            # L_along_curve / L = theta / 2pi
            lengths = np.sqrt(
                np.diff(closedX) ** 2 + np.diff(closedY) ** 2 + np.diff(closedZ) ** 2
            )
            thetas = 2 * np.pi * np.cumsum(lengths) / np.sum(lengths)
            thetas = np.insert(thetas, 0, 0)
            s = thetas[:-1]
        elif s is None:
            s = np.linspace(0, 2 * np.pi, X.size, endpoint=False)
        else:
            s = np.atleast_1d(s)
            errorif(
                not np.all(np.diff(s) > 0),
                ValueError,
                "supplied s must be monotonically increasing!",
            )
            errorif(s[0] < 0, ValueError, "s must lie in [0, 2pi]")
            errorif(s[-1] > 2 * np.pi, ValueError, "s must lie in [0, 2pi]")

        grid = LinearGrid(zeta=s, NFP=1, sym=False)
        basis = FourierSeries(N=N, NFP=1, sym=False)
        transform = Transform(grid, basis, build_pinv=True)
        X_n = transform.fit(coords_xyz[:, 0])
        Y_n = transform.fit(coords_xyz[:, 1])
        Z_n = transform.fit(coords_xyz[:, 2])
        return FourierXYZCurve(X_n=X_n, Y_n=Y_n, Z_n=Z_n, name=name)


class FourierPlanarCurve(Curve):
    """Curve that lies in a plane.

    Parameterized by a point (the center of the curve), a vector (normal to the plane),
    and a Fourier series defining the radius from the center as a function of
    a polar angle theta.

    Parameters
    ----------
    center : array-like, shape(3,)
        x,y,z coordinates of center of curve
    normal : array-like, shape(3,)
        x,y,z components of normal vector to planar surface
    r_n : array-like
        Fourier coefficients for radius from center as function of polar angle
    modes : array-like
        mode numbers associated with r_n
    name : str
        name for this curve

    """

    _io_attrs_ = Curve._io_attrs_ + [
        "_r_n",
        "_center",
        "_normal",
        "_r_basis",
    ]

    # Reference frame is centered at the origin with normal in the +Z direction.
    # The curve is computed in this frame and then shifted/rotated to the correct frame.
    def __init__(
        self,
        center=[10, 0, 0],
        normal=[0, 1, 0],
        r_n=2,
        modes=None,
        name="",
    ):
        super().__init__(name)
        r_n = np.atleast_1d(r_n)
        if modes is None:
            modes = np.arange(-(r_n.size // 2), r_n.size // 2 + 1)
        else:
            modes = np.asarray(modes)
        assert issubclass(modes.dtype.type, np.integer)

        N = np.max(abs(modes))
        self._r_basis = FourierSeries(N, NFP=1, sym=False)
        self._r_n = copy_coeffs(r_n, modes, self.r_basis.modes[:, 2])

        self.normal = normal
        self.center = center

    @property
    def r_basis(self):
        """Spectral basis for Fourier series."""
        return self._r_basis

    @property
    def N(self):
        """Maximum mode number."""
        return self.r_basis.N

    def change_resolution(self, N=None):
        """Change the maximum angular resolution."""
        if (N is not None) and (N != self.N):
            modes_old = self.r_basis.modes
            self.r_basis.change_resolution(N=N)
            self.r_n = copy_coeffs(self.r_n, modes_old, self.r_basis.modes)

    @property
    def center(self):
        """Center of planar curve polar coordinates."""
        return self._center

    @center.setter
    def center(self, new):
        if len(new) == 3:
            self._center = np.asarray(new)
        else:
            raise ValueError(
                "center should be a 3 element vector [cx, cy, cz], got {}".format(new)
            )

    @property
    def normal(self):
        """Normal vector to plane."""
        return self._normal

    @normal.setter
    def normal(self, new):
        if len(np.asarray(new)) == 3:
            self._normal = np.asarray(new) / np.linalg.norm(new)
        else:
            raise ValueError(
                "normal should be a 3 element vector [nx, ny, nz], got {}".format(new)
            )

    @property
    def r_n(self):
        """Spectral coefficients for r."""
        return self._r_n

    @r_n.setter
    def r_n(self, new):
        if len(np.asarray(new)) == self.r_basis.num_modes:
            self._r_n = jnp.asarray(new)
        else:
            raise ValueError(
                f"r_n should have the same size as the basis, got {len(new)} for "
                + f"basis with {self.r_basis.num_modes} modes."
            )

    def get_coeffs(self, n):
        """Get Fourier coefficients for given mode number(s)."""
        n = np.atleast_1d(n).astype(int)
        r = np.zeros_like(n).astype(float)

        idx = np.where(n[:, np.newaxis] == self.r_basis.modes[:, 2])

        r[idx[0]] = self.r_n[idx[1]]
        return r

    def set_coeffs(self, n, r=None):
        """Set specific Fourier coefficients."""
        n, r = np.atleast_1d(n), np.atleast_1d(r)
        r = np.broadcast_to(r, n.shape)
        for nn, rr in zip(n, r):
            idx = self.r_basis.get_idx(0, 0, nn)
            if rr is not None:
                self.r_n = put(self.r_n, idx, rr)


class SplineXYZCurve(Curve):
    """Curve parameterized by spline knots in X,Y,Z.

    Parameters
    ----------
    X, Y, Z: array-like
        Points for X, Y, Z describing the curve. If the endpoint is included
        (ie, X[0] == X[-1]), then the final point will be dropped.
    knots : ndarray or "arclength"
        arbitrary curve parameter values to use for spline knots,
        should be a monotonic, 1D ndarray of same length as the input X,Y,Z.
        If None, defaults to using an linearly spaced points in [0, 2pi) as the knots.
        If supplied, should lie in [0,2pi].
        Alternatively, the string "arclength" can be supplied to use the normalized
        distance between points.
    method : str
        method of interpolation

        - ``'nearest'``: nearest neighbor interpolation
        - ``'linear'``: linear interpolation
        - ``'cubic'``: C1 cubic splines (aka local splines)
        - ``'cubic2'``: C2 cubic splines (aka natural splines)
        - ``'catmull-rom'``: C1 cubic centripetal "tension" splines
        - ``'cardinal'``: C1 cubic general tension splines. If used, default tension of
          c = 0 will be used
        - ``'monotonic'``: C1 cubic splines that attempt to preserve monotonicity in the
          data, and will not introduce new extrema in the interpolated points
        - ``'monotonic-0'``: same as `'monotonic'` but with 0 first derivatives at both
          endpoints

    name : str
        name for this curve

    """

    _io_attrs_ = Curve._io_attrs_ + ["_X", "_Y", "_Z", "_knots", "_method"]

    def __init__(
        self,
        X,
        Y,
        Z,
        knots=None,
        method="cubic",
        name="",
    ):
        super().__init__(name)
        X, Y, Z = np.atleast_1d(X), np.atleast_1d(Y), np.atleast_1d(Z)
        X, Y, Z = np.broadcast_arrays(X, Y, Z)

        X, Y, Z, closedX, closedY, closedZ, closed_flag = _unclose_curve(X, Y, Z)

        self._X = X
        self._Y = Y
        self._Z = Z

        if isinstance(knots, str):
            assert knots == "arclength", f"got unknown arclength specification {knots}"
            # find equal arclength angle-like variable, and use that as theta
            # L_along_curve / L = theta / 2pi
            lengths = np.sqrt(
                np.diff(closedX) ** 2 + np.diff(closedY) ** 2 + np.diff(closedZ) ** 2
            )
            thetas = 2 * np.pi * np.cumsum(lengths) / np.sum(lengths)
            thetas = np.insert(thetas, 0, 0)
            knots = thetas[:-1]
        elif knots is None:
            knots = np.linspace(0, 2 * np.pi, len(self._X), endpoint=False)
        else:
            knots = np.atleast_1d(knots)
            errorif(
                not np.all(np.diff(knots) > 0),
                ValueError,
                "supplied knots must be monotonically increasing!",
            )
            errorif(knots[0] < 0, ValueError, "knots must lie in [0, 2pi]")
            errorif(knots[-1] > 2 * np.pi, ValueError, "knots must lie in [0, 2pi]")
            knots = knots[:-1] if closed_flag else knots

        self._knots = knots
        self.method = method

    @property
    def X(self):
        """Coordinates for X."""
        return self._X

    @X.setter
    def X(self, new):
        if len(new) == len(self.knots):
            self._X = jnp.asarray(new)
        else:
            raise ValueError(
                "X should have the same size as the knots, "
                + f"got {len(new)} X values for {len(self.knots)} knots"
            )

    @property
    def Y(self):
        """Coordinates for Y."""
        return self._Y

    @Y.setter
    def Y(self, new):
        if len(new) == len(self.knots):
            self._Y = jnp.asarray(new)
        else:
            raise ValueError(
                "Y should have the same size as the knots, "
                + f"got {len(new)} Y values for {len(self.knots)} knots"
            )

    @property
    def Z(self):
        """Coordinates for Z."""
        return self._Z

    @Z.setter
    def Z(self, new):
        if len(new) == len(self.knots):
            self._Z = jnp.asarray(new)
        else:
            raise ValueError(
                "Z should have the same size as the knots, "
                + f"got {len(new)} Z values for {len(self.knots)} knots"
            )

    @property
    def knots(self):
        """Knots for spline."""
        return self._knots

    @knots.setter
    def knots(self, new):
        if len(new) == len(self.knots):
            knots = jnp.atleast_1d(new)
            errorif(
                not np.all(np.diff(knots) > 0),
                ValueError,
                "supplied knots must be monotonically increasing!",
            )
            errorif(knots[0] < 0, ValueError, "knots must lie in [0, 2pi]")
            errorif(knots[-1] > 2 * np.pi, ValueError, "knots must lie in [0, 2pi]")
            self._knots = jnp.asarray(knots)
        else:
            raise ValueError(
                "new knots should have the same size as the current knots, "
                + f"got {len(new)} new knots, but expected {len(self.knots)} knots"
            )

    @property
    def N(self):
        """Number of knots in the spline."""
        return self.knots.size

    @property
    def method(self):
        """Method of interpolation to use."""
        return self._method

    @method.setter
    def method(self, new):
        possible_methods = [
            "nearest",
            "linear",
            "cubic",
            "cubic2",
            "catmull-rom",
            "monotonic",
            "monotonic-0",
            "cardinal",
        ]
        if new in possible_methods:
            self._method = new
        else:
            raise ValueError(
                "Method must be one of {possible_methods}, "
                + f"instead got unknown method {new} "
            )

    @classmethod
    def from_values(cls, coords, knots=None, method="cubic", name="", basis="xyz"):
        """Create SplineXYZCurve from coordinate values.

        Parameters
        ----------
        coords: ndarray
            Points for X, Y, Z describing the curve. If the endpoint is included
            (ie, X[0] == X[-1]), then the final point will be dropped.
        knots : ndarray
            arbitrary curve parameter values to use for spline knots,
            should be an 1D ndarray of same length as the input.
            (input length in this case is determined by grid argument, since
            the input coordinates come from
            Curve.compute("x",grid=grid))
            If None, defaults to using an equal-arclength angle as the knots
            If supplied, will be rescaled to lie in [0,2pi]
        method : str
            method of interpolation

            - `'nearest'`: nearest neighbor interpolation
            - `'linear'`: linear interpolation
            - `'cubic'`: C1 cubic splines (aka local splines)
            - `'cubic2'`: C2 cubic splines (aka natural splines)
            - `'catmull-rom'`: C1 cubic centripetal "tension" splines

        name : str
            name for this curve
        basis : {"rpz", "xyz"}
            basis for input coordinates. Defaults to "xyz"

        Returns
        -------
        curve: SplineXYZCurve
            New representation of the curve parameterized by splines in X,Y,Z.

        """
        if basis == "rpz":
            coords = rpz2xyz(coords)
        return SplineXYZCurve(
            coords[:, 0], coords[:, 1], coords[:, 2], knots, method, name
        )



"""Wrappers for doing STELLOPT/SIMSOPT like optimization."""

import numpy as np

from desc.backend import jnp
from desc.compute import arg_order
from desc.objectives import (
    BoundaryRSelfConsistency,
    BoundaryZSelfConsistency,
    ObjectiveFunction,
    get_fixed_boundary_constraints,
    maybe_add_self_consistency,
)
from desc.objectives.utils import align_jacobian, factorize_linear_constraints
from desc.utils import Timer, get_instance

from .utils import compute_jac_scale, f_where_x


class LinearConstraintProjection(ObjectiveFunction):
    """Remove linear constraints via orthogonal projection.

    Given a problem of the form

    min_x f(x)  subject to A*x=b

    We can write any feasible x=xp + Z*x_reduced where xp is a particular solution to
    Ax=b (taken to be the least norm solution), Z is a representation for the null
    space of A (A*Z=0) and x_reduced is unconstrained. This transforms the problem into

    min_x_reduced f(x_reduced)

    Parameters
    ----------
    objective : ObjectiveFunction
        Objective function to optimize.
    constraints : tuple of Objective
        Linear constraints to enforce. Should be a tuple or list of Objective,
        and must all be linear.
    eq : Equilibrium, optional
        Equilibrium that will be optimized to satisfy the objectives.
    verbose : int, optional
        Level of output.
    """

    def __init__(self, objective, constraints, eq=None, verbose=1):
        assert isinstance(objective, ObjectiveFunction), (
            "objective should be instance of ObjectiveFunction." ""
        )
        for con in constraints:
            if not con.linear:
                raise ValueError(
                    "LinearConstraintProjection method "
                    + "cannot handle nonlinear constraint {}.".format(con)
                )
        self._objective = objective
        self._objectives = [objective]
        self._constraints = constraints
        self._built = False
        # don't want to compile this, just use the compiled objective
        self._use_jit = False
        self._compiled = False
        self._eq = eq

    def build(self, eq=None, use_jit=None, verbose=1):
        """Build the objective.

        Parameters
        ----------
        eq : Equilibrium, optional
            Equilibrium that will be optimized to satisfy the Objective.
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
            Note: unused by this class, should pass to sub-objectives directly.
        verbose : int, optional
            Level of output.

        """
        eq = eq or self._eq
        timer = Timer()
        timer.start("Linear constraint projection build")

        # we don't always build here because in ~all cases the user doesn't interact
        # with this directly, so if the user wants to manually rebuild they should
        # do it before this wrapper is created for them.
        if not self._objective.built:
            self._objective.build(eq, verbose=verbose)
        for con in self._constraints:
            if not con.built:
                con.build(eq, verbose=verbose)

        args = np.concatenate([obj.args for obj in self._constraints])
        args = np.concatenate((args, self._objective.args))
        # this is all args used by both constraints and objective
        self._args = [arg for arg in arg_order if arg in args]
        self._dim_f = self._objective.dim_f
        self._dimensions = self._objective.dimensions
        self._scalar = self._objective.scalar
        self._dimensions = self._objective.dimensions
        self._objective.set_args(*self._args)
        self.set_args()
        (
            self._xp,
            self._A,
            self._b,
            self._Z,
            self._unfixed_idx,
            self._project,
            self._recover,
        ) = factorize_linear_constraints(
            self._constraints,
            self._args,
        )
        self._dim_x_full = self._dim_x
        self._dim_x = self._Z.shape[1]

        self._built = True
        timer.stop("Linear constraint projection build")
        if verbose > 1:
            timer.disp("Linear constraint projection build")

    def compile(self, mode="lsq", verbose=1):
        """Call the necessary functions to ensure the function is compiled.

        Parameters
        ----------
        mode : {"auto", "lsq", "scalar", "all"}
            Whether to compile for least squares optimization or scalar optimization.
            "auto" compiles based on the type of objective,
            "all" compiles all derivatives.
        verbose : int, optional
            Level of output.

        """
        self._objective.compile(mode, verbose)

    def project(self, x):
        """Project full vector x into x_reduced that satisfies constraints."""
        return self._project(x)

    def recover(self, x_reduced):
        """Recover the full state vector from the reduced optimization vector."""
        return self._recover(x_reduced)

    def x(self, eq):
        """Return the reduced state vector from the Equilibrium eq."""
        x = np.zeros((self._dim_x_full,))
        for arg in self.args:
            x[self.x_idx[arg]] = getattr(eq, arg)
        return self.project(x)

    def unpack_state(self, x):
        """Unpack the state vector into its components.

        Parameters
        ----------
        x : ndarray
            State vector, either full or reduced.

        Returns
        -------
        kwargs : dict
            Dictionary of the state components with argument names as keys.

        """
        if x.size != self.dim_x:
            raise ValueError(
                "Input vector dimension is invalid, expected "
                + f"{self.dim_x} got {x.size}."
            )
        x = self.recover(x)
        return self._objective.unpack_state(x)

    def compute_unscaled(self, x_reduced, constants=None):
        """Compute the unscaled form of the objective function.

        Parameters
        ----------
        x_reduced : ndarray
            Reduced state vector that satisfies linear constraints.
        constants : list
            Constant parameters passed to sub-objectives.

        Returns
        -------
        f : ndarray
            Objective function value(s).

        """
        x = self.recover(x_reduced)
        f = self._objective.compute_unscaled(x, constants)
        return f

    def compute_scaled(self, x_reduced, constants=None):
        """Compute the objective function and apply weighting / normalization.

        Parameters
        ----------
        x_reduced : ndarray
            Reduced state vector that satisfies linear constraints.
        constants : list
            Constant parameters passed to sub-objectives.

        Returns
        -------
        f : ndarray
            Objective function value(s).

        """
        x = self.recover(x_reduced)
        f = self._objective.compute_scaled(x, constants)
        return f

    def compute_scaled_error(self, x_reduced, constants=None):
        """Compute the objective function and apply weighting / bounds.

        Parameters
        ----------
        x_reduced : ndarray
            Reduced state vector that satisfies linear constraints.
        constants : list
            Constant parameters passed to sub-objectives.

        Returns
        -------
        f : ndarray
            Objective function value(s).

        """
        x = self.recover(x_reduced)
        f = self._objective.compute_scaled_error(x, constants)
        return f

    def compute_scalar(self, x_reduced, constants=None):
        """Compute the scalar form of the objective function.

        Parameters
        ----------
        x_reduced : ndarray
            Reduced state vector that satisfies linear constraints.
        constants : list
            Constant parameters passed to sub-objectives.

        Returns
        -------
        f : float
            Objective function value.

        """
        x = self.recover(x_reduced)
        return self._objective.compute_scalar(x, constants)

    def grad(self, x_reduced, constants=None):
        """Compute gradient of the sum of squares of residuals.

        Parameters
        ----------
        x_reduced : ndarray
            Reduced state vector that satisfies linear constraints.
        constants : list
            Constant parameters passed to sub-objectives.

        Returns
        -------
        g : ndarray
            gradient vector.
        """
        x = self.recover(x_reduced)
        df = self._objective.grad(x, constants)
        return df[self._unfixed_idx] @ self._Z

    def hess(self, x_reduced, constants=None):
        """Compute Hessian of the sum of squares of residuals.

        Parameters
        ----------
        x_reduced : ndarray
            Reduced state vector that satisfies linear constraints.
        constants : list
            Constant parameters passed to sub-objectives.

        Returns
        -------
        H : ndarray
            Hessian matrix.
        """
        x = self.recover(x_reduced)
        df = self._objective.hess(x, constants)
        return self._Z.T @ df[self._unfixed_idx, :][:, self._unfixed_idx] @ self._Z

    def jac_unscaled(self, x_reduced, constants):
        """Compute Jacobian of the vector objective function without weighting / bounds.

        Parameters
        ----------
        x_reduced : ndarray
            Reduced state vector that satisfies linear constraints.
        constants : list
            Constant parameters passed to sub-objectives.

        Returns
        -------
        J : ndarray
            Jacobian matrix.
        """
        x = self.recover(x_reduced)
        df = self._objective.jac_unscaled(x, constants)
        return df[:, self._unfixed_idx] @ self._Z

    def jac_scaled(self, x_reduced, constants=None):
        """Compute Jacobian of the vector objective function with weighting / bounds.

        Parameters
        ----------
        x_reduced : ndarray
            Reduced state vector that satisfies linear constraints.
        constants : list
            Constant parameters passed to sub-objectives.

        Returns
        -------
        J : ndarray
            Jacobian matrix.
        """
        x = self.recover(x_reduced)
        df = self._objective.jac_scaled(x, constants)
        return df[:, self._unfixed_idx] @ self._Z

    def vjp_scaled(self, v, x_reduced):
        """Compute vector-Jacobian product of the objective function.

        Uses the scaled form of the objective.

        Parameters
        ----------
        v : ndarray
            Vector to left-multiply the Jacobian by.
        x_reduced : ndarray
            Optimization variables.

        """
        x = self.recover(x_reduced)
        df = self._objective.vjp_scaled(v, x)
        return df[self._unfixed_idx] @ self._Z

    def vjp_unscaled(self, v, x_reduced):
        """Compute vector-Jacobian product of the objective function.

        Uses the unscaled form of the objective.

        Parameters
        ----------
        v : ndarray
            Vector to left-multiply the Jacobian by.
        x_reduced : ndarray
            Optimization variables.

        """
        x = self.recover(x_reduced)
        df = self._objective.vjp_unscaled(v, x)
        return df[self._unfixed_idx] @ self._Z

    @property
    def constants(self):
        """list: constant parameters for each sub-objective."""
        return self._objective.constants

    @property
    def target_scaled(self):
        """ndarray: target vector."""
        return self._objective.target_scaled

    @property
    def bounds_scaled(self):
        """tuple: lower and upper bounds for residual vector."""
        return self._objective.bounds_scaled

    @property
    def weights(self):
        """ndarray: weight vector."""
        return self._objective.weights


class ProximalProjection(ObjectiveFunction):
    """Remove equilibrium constraint by projecting onto constraint at each step.

    Combines objective and equilibrium constraint into a single objective to then pass
    to an unconstrained optimizer.

    At each iteration, after a step is taken to reduce the objective, the equilibrium
    is perturbed and re-solved to bring it back into force balance. This is analogous
    to a proximal method where each iterate is projected back onto the feasible set.

    Parameters
    ----------
    objective : ObjectiveFunction
        Objective function to optimize.
    constraint : ObjectiveFunction
        Equilibrium constraint to enforce. Should be an ObjectiveFunction with one or
        more of the following objectives: {ForceBalance, CurrentDensity,
        RadialForceBalance, HelicalForceBalance}
    eq : Equilibrium, optional
        Equilibrium that will be optimized to satisfy the objectives.
    verbose : int, optional
        Level of output.
    perturb_options, solve_options : dict
        dictionary of arguments passed to Equilibrium.perturb and Equilibrium.solve
        during the projection step.
    """

    def __init__(
        self,
        objective,
        constraint,
        eq=None,
        verbose=1,
        perturb_options=None,
        solve_options=None,
    ):
        assert isinstance(objective, ObjectiveFunction), (
            "objective should be instance of ObjectiveFunction." ""
        )
        assert isinstance(constraint, ObjectiveFunction), (
            "constraint should be instance of ObjectiveFunction." ""
        )
        for con in constraint.objectives:
            if not con._equilibrium:
                raise ValueError(
                    "ProximalProjection method "
                    + "cannot handle general nonlinear constraint {}.".format(con)
                )
        self._objective = objective
        self._constraint = constraint
        solve_options = {} if solve_options is None else solve_options
        perturb_options = {} if perturb_options is None else perturb_options
        self._objectives = [objective, constraint]
        perturb_options.setdefault("verbose", 0)
        perturb_options.setdefault("include_f", False)
        solve_options.setdefault("verbose", 0)
        self._perturb_options = perturb_options
        self._solve_options = solve_options
        self._built = False
        # don't want to compile this, just use the compiled objective and constraint
        self._use_jit = False
        self._compiled = False
        self._eq = eq

    def set_args(self, *args):
        """Set which arguments the objective should expect.

        Defaults to args from all sub-objectives. Additional arguments can be passed in.
        """
        # this is everything taken by either objective
        self._full_args = (
            self._constraint.args
            + self._objective.args
            + list(np.concatenate([obj.args for obj in self._linear_constraints]))
        )
        self._full_args = [arg for arg in arg_order if arg in self._full_args]

        # arguments being optimized are all args, but with internal degrees of freedom
        # replaced by boundary terms
        self._args = self._full_args.copy() + list(args)
        self._args = [arg for arg in arg_order if arg in self._args]
        if "L_lmn" in self._args:
            self._args.remove("L_lmn")
        if "Ra_n" in self._args:
            self._args.remove("Ra_n")
        if "Za_n" in self._args:
            self._args.remove("Za_n")
        if "R_lmn" in self._args:
            self._args.remove("R_lmn")
            if "Rb_lmn" not in self._args:
                self._args.append("Rb_lmn")
        if "Z_lmn" in self._args:
            self._args.remove("Z_lmn")
            if "Zb_lmn" not in self._args:
                self._args.append("Zb_lmn")
        self._set_state_vector()

    def _set_state_vector(self):
        self._dimensions = self._objective.dimensions
        self._dim_x = 0
        self._x_idx = {}
        for arg in self.args:
            self.x_idx[arg] = np.arange(self._dim_x, self._dim_x + self.dimensions[arg])
            self._dim_x += self.dimensions[arg]
        self._dim_x_full = 0
        self._x_idx_full = {}
        for arg in self._full_args:
            self._x_idx_full[arg] = np.arange(
                self._dim_x_full, self._dim_x_full + self.dimensions[arg]
            )
            self._dim_x_full += self.dimensions[arg]

        (
            xp,
            A,
            b,
            self._Z,
            self._unfixed_idx,
            project,
            recover,
        ) = factorize_linear_constraints(self._linear_constraints, self._full_args)

        # dx/dc - goes from the full state to optimization variables
        x_idx = np.concatenate(
            [
                self._x_idx_full[arg]
                for arg in self._args
                if arg not in ["Rb_lmn", "Zb_lmn"]
            ]
        )
        x_idx.sort(kind="mergesort")
        self._dxdc = np.eye(self._dim_x_full)[:, x_idx]
        if "Rb_lmn" in self._args:
            c = get_instance(self._linear_constraints, BoundaryRSelfConsistency)
            A = c.derivatives["jac_unscaled"]["R_lmn"](
                *[jnp.zeros(c.dimensions[arg]) for arg in c.args]
            )
            Ainv = np.linalg.pinv(A)
            dxdRb = np.eye(self._dim_x_full)[:, self._x_idx_full["R_lmn"]] @ Ainv
            self._dxdc = np.hstack((self._dxdc, dxdRb))
        if "Zb_lmn" in self._args:
            c = get_instance(self._linear_constraints, BoundaryZSelfConsistency)
            A = c.derivatives["jac_unscaled"]["Z_lmn"](
                *[jnp.zeros(c.dimensions[arg]) for arg in c.args]
            )
            Ainv = np.linalg.pinv(A)
            dxdZb = np.eye(self._dim_x_full)[:, self._x_idx_full["Z_lmn"]] @ Ainv
            self._dxdc = np.hstack((self._dxdc, dxdZb))

    def build(self, eq=None, use_jit=None, verbose=1):  # noqa: C901
        """Build the objective.

        Parameters
        ----------
        eq : Equilibrium, optional
            Equilibrium that will be optimized to satisfy the Objective.
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
            Note: unused by this class, should pass to sub-objectives directly.
        verbose : int, optional
            Level of output.

        """
        eq = eq or self._eq
        timer = Timer()
        timer.start("Proximal projection build")

        self._eq = eq.copy()
        self._linear_constraints = get_fixed_boundary_constraints(
            eq=eq,
            iota=self._eq.iota is not None,
            kinetic=self._eq.electron_temperature is not None,
        )
        self._linear_constraints = maybe_add_self_consistency(
            eq, self._linear_constraints
        )

        # we don't always build here because in ~all cases the user doesn't interact
        # with this directly, so if the user wants to manually rebuild they should
        # do it before this wrapper is created for them.
        if not self._objective.built:
            self._objective.build(self._eq, verbose=verbose)
        if not self._constraint.built:
            self._constraint.build(self._eq, verbose=verbose)

        for constraint in self._linear_constraints:
            constraint.build(self._eq, verbose=verbose)

        self._dim_f = self._objective.dim_f
        if self._dim_f == 1:
            self._scalar = True
        else:
            self._scalar = False

        self.set_args()
        self._objective.set_args(*self._full_args)
        self._constraint.set_args(*self._full_args)
        # history and caching
        self._x_old = np.zeros((self._dim_x,))
        for arg in self.args:
            self._x_old[self.x_idx[arg]] = getattr(eq, arg)

        self._allx = [self._x_old]
        self._allxopt = [self._objective.x(eq)]
        self._allxeq = [self._constraint.x(eq)]
        self.history = {}
        for arg in arg_order:
            self.history[arg] = [np.asarray(getattr(self._eq, arg)).copy()]

        self._built = True
        timer.stop("Proximal projection build")
        if verbose > 1:
            timer.disp("Proximal projection build")

    def compile(self, mode="lsq", verbose=1):
        """Call the necessary functions to ensure the function is compiled.

        Parameters
        ----------
        mode : {"auto", "lsq", "scalar", "all"}
            Whether to compile for least squares optimization or scalar optimization.
            "auto" compiles based on the type of objective,
            "all" compiles all derivatives.
        verbose : int, optional
            Level of output.

        """
        self._objective.compile(mode, verbose)
        self._constraint.compile(mode, verbose)

    def _update_equilibrium(self, x, store=False):
        """Update the internal equilibrium with new boundary, profile etc.

        Parameters
        ----------
        x : ndarray
            New values of optimization variables.
        store : bool
            Whether the new x should be stored in self.history

        Notes
        -----
        After updating, if store=False, self._eq will revert back to the previous
        solution when store was True
        """
        # first check if its something we've seen before, if it is just return
        # cached value, no need to perturb + resolve
        xopt = f_where_x(x, self._allx, self._allxopt)
        xeq = f_where_x(x, self._allx, self._allxeq)
        if xopt.size > 0 and xeq.size > 0:
            pass
        else:
            x_dict = self.unpack_state(x)
            x_dict_old = self.unpack_state(self._x_old)
            deltas = {str(key): x_dict[key] - x_dict_old[key] for key in x_dict}
            self._eq = self._eq.perturb(
                objective=self._constraint,
                constraints=self._linear_constraints,
                deltas=deltas,
                **self._perturb_options,
            )
            self._eq.solve(
                objective=self._constraint,
                constraints=self._linear_constraints,
                **self._solve_options,
            )
            xopt = self._objective.x(self._eq)
            xeq = self._constraint.x(self._eq)
            self._allx.append(x)
            self._allxopt.append(xopt)
            self._allxeq.append(xeq)

        if store:
            self._x_old = x
            xd = self.unpack_state(x)
            xod = self._objective.unpack_state(xopt)
            xed = self._constraint.unpack_state(xeq)
            xd.update(xod)
            xd.update(xed)
            for arg in arg_order:
                val = xd.get(arg, self.history[arg][-1])
                self.history[arg] += [np.asarray(val).copy()]
                # ensure eq has correct values if we didn't go into else block above.
                if val.size:
                    setattr(self._eq, arg, val)
            for con in self._linear_constraints:
                if hasattr(con, "update_target"):
                    con.update_target(self._eq)
        else:
            for arg in arg_order:
                val = self.history[arg][-1].copy()
                if val.size:
                    setattr(self._eq, arg, val)
            for con in self._linear_constraints:
                if hasattr(con, "update_target"):
                    con.update_target(self._eq)
        return xopt, xeq

    def compute_scaled(self, x, constants=None):
        """Compute the objective function and apply weights/normalization.

        Parameters
        ----------
        x : ndarray
            State vector.
        constants : list
            Constant parameters passed to sub-objectives.

        Returns
        -------
        f : ndarray
            Objective function value(s).

        """
        if constants is None:
            constants = self.constants
        xopt, _ = self._update_equilibrium(x, store=False)
        return self._objective.compute_scaled(xopt, constants[0])

    def compute_scaled_error(self, x, constants=None):
        """Compute the error between target and objective and apply weights etc.

        Parameters
        ----------
        x : ndarray
            State vector.
        constants : list
            Constant parameters passed to sub-objectives.

        Returns
        -------
        f : ndarray
            Objective function value(s).

        """
        if constants is None:
            constants = self.constants
        xopt, _ = self._update_equilibrium(x, store=False)
        return self._objective.compute_scaled_error(xopt, constants[0])

    def compute_unscaled(self, x, constants=None):
        """Compute the raw value of the objective function.

        Parameters
        ----------
        x : ndarray
            State vector.
        constants : list
            Constant parameters passed to sub-objectives.

        Returns
        -------
        f : ndarray
            Objective function value(s).

        """
        if constants is None:
            constants = self.constants
        xopt, _ = self._update_equilibrium(x, store=False)
        return self._objective.compute_unscaled(xopt, constants[0])

    def grad(self, x, constants=None):
        """Compute gradient of the sum of squares of residuals.

        Parameters
        ----------
        x : ndarray
            State vector.
        constants : list
            Constant parameters passed to sub-objectives.

        Returns
        -------
        g : ndarray
            gradient vector.
        """
        f = jnp.atleast_1d(self.compute_scaled_error(x, constants))
        J = self.jac_scaled(x, constants)
        return f.T @ J

    def jac_unscaled(self, x, constants=None):
        """Compute Jacobian of the vector objective function without weights / bounds.

        Parameters
        ----------
        x : ndarray
            State vector.
        constants : list
            Constant parameters passed to sub-objectives.

        Returns
        -------
        J : ndarray
            Jacobian matrix.
        """
        raise NotImplementedError("Unscaled jacobian of proximal projection is hard.")

    def jac_scaled(self, x, constants=None):
        """Compute Jacobian of the vector objective function with weights / bounds.

        Parameters
        ----------
        x : ndarray
            State vector.
        constants : list
            Constant parameters passed to sub-objectives.

        Returns
        -------
        J : ndarray
            Jacobian matrix.
        """
        if constants is None:
            constants = self.constants
        xg, xf = self._update_equilibrium(x, store=True)

        # Jacobian matrices wrt combined state vectors
        Fx = self._constraint.jac_scaled(xf, constants[1])
        Gx = self._objective.jac_scaled(xg, constants[0])
        Fx = align_jacobian(Fx, self._constraint, self._objective)
        Gx = align_jacobian(Gx, self._objective, self._constraint)

        # projections onto optimization space
        # possibly better way: Gx @ np.eye(Gx.shape[1])[:,self._unfixed_idx] @ self._Z
        Fx_reduced = Fx[:, self._unfixed_idx] @ self._Z
        Gx_reduced = Gx[:, self._unfixed_idx] @ self._Z
        Fc = Fx @ self._dxdc
        Gc = Gx @ self._dxdc

        # some scaling to improve conditioning
        wf, _ = compute_jac_scale(Fx_reduced)
        wg, _ = compute_jac_scale(Gx_reduced)
        wx = wf + wg
        Fxh = Fx_reduced * wx
        Gxh = Gx_reduced * wx

        cutoff = np.finfo(Fxh.dtype).eps * np.max(Fxh.shape)
        uf, sf, vtf = np.linalg.svd(Fxh, full_matrices=False)
        sf += sf[-1]  # add a tiny bit of regularization
        sfi = np.where(sf < cutoff * sf[0], 0, 1 / sf)
        Fxh_inv = vtf.T @ (sfi[..., np.newaxis] * uf.T)

        # TODO: make this more efficient for finite differences etc. Can probably
        # reduce the number of operations and tangents
        LHS = Gxh @ (Fxh_inv @ Fc) - Gc
        return -LHS

    def hess(self, x, constants=None):
        """Compute Hessian of the sum of squares of residuals.

        Uses the "small residual approximation" where the Hessian is replaced by
        the square of the Jacobian: H = J.T @ J

        Parameters
        ----------
        x : ndarray
            State vector.
        constants : list
            Constant parameters passed to sub-objectives.

        Returns
        -------
        H : ndarray
            Hessian matrix.
        """
        J = self.jac_scaled(x, constants)
        return J.T @ J

    def vjp_scaled(self, v, x):
        """Compute vector-Jacobian product of the objective function.

        Uses the scaled form of the objective.

        Parameters
        ----------
        v : ndarray
            Vector to left-multiply the Jacobian by.
        x : ndarray
            Optimization variables.

        """
        raise NotImplementedError

    def vjp_unscaled(self, v, x):
        """Compute vector-Jacobian product of the objective function.

        Uses the unscaled form of the objective.

        Parameters
        ----------
        v : ndarray
            Vector to left-multiply the Jacobian by.
        x : ndarray
            Optimization variables.

        """
        raise NotImplementedError

    @property
    def constants(self):
        """list: constant parameters for each sub-objective."""
        return [self._objective.constants, self._constraint.constants]

    @property
    def target_scaled(self):
        """ndarray: target vector."""
        return self._objective.target_scaled

    @property
    def bounds_scaled(self):
        """tuple: lower and upper bounds for residual vector."""
        return self._objective.bounds_scaled

    @property
    def weights(self):
        """ndarray: weight vector."""
        return self._objective.weights
